from typing import Annotated, overload

import numpy
from numpy.typing import NDArray

import nwsspc.sharp.calc.layer


class WindVector:
    @overload
    def __init__(self) -> None: ...

    @overload
    def __init__(self, speed: float, direction: float) -> None: ...

    @property
    def speed(self) -> float:
        """Wind Speed (m/s)"""

    @speed.setter
    def speed(self, arg: float, /) -> None: ...

    @property
    def direction(self) -> float:
        """Wind Direction (degrees from North)"""

    @direction.setter
    def direction(self, arg: float, /) -> None: ...

class WindComponents:
    @overload
    def __init__(self) -> None: ...

    @overload
    def __init__(self, u_comp: float, v_comp: float) -> None: ...

    @property
    def u(self) -> float:
        """U wind component (m/s)"""

    @u.setter
    def u(self, arg: float, /) -> None: ...

    @property
    def v(self) -> float:
        """V wind component (m/s)"""

    @v.setter
    def v(self, arg: float, /) -> None: ...

@overload
def helicity(layer: nwsspc.sharp.calc.layer.HeightLayer, storm_motion: WindComponents, height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], u_wind: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], v_wind: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> float:
    """
    Computes the Storm Relative Helicity (SRH) over a given layer using 
    storm motion vector components stored in a WindComponents object.

    This integration occurs over the given arrays, using interpolation 
    for the top and bottom of the integration layer, and native data levels
    in between. 

    Parameters
    ----------
    layer : nwsspc.sharp.calc.layer.HeightLayer 
        A HeightLayer to integrate over (meters AGL)
    storm_motion : nwsspc.sharp.calc.winds.WindComponents 
        A WindComponents object with the storm motion in m/s 
    height : numpy.ndarray[dtype=float32] 
        1D NumPy array of height values (meters)
    u_wind : numpy.ndarray[dtype=float32]
        1D NumPy array of environment U-wind components (m/s)
    v_wind : numpy.ndarray[dtype=float32]
        1D NumPy array of environment V-wind components (m/s)

    Returns
    -------
    float
        Storm Relative Helicity (m^2/s^2)
    """

@overload
def helicity(layer: nwsspc.sharp.calc.layer.PressureLayer, storm_motion: WindComponents, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], u_wind: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], v_wind: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> float:
    """
    Computes the Storm Relative Helicity (SRH) over a given layer using 
    storm motion vector components stored in a WindComponents object.

    This integration occurs over the given arrays, using interpolation 
    for the top and bottom of the integration layer, and native data levels
    in between. 

    Parameters
    ----------
    layer : nwsspc.sharp.calc.layer.PressureLayer 
        A PressureLayer to integrate over (Pa)
    storm_motion : nwsspc.sharp.calc.winds.WindComponents 
        A WindComponents object with the storm motion in m/s 
    height : numpy.ndarray[dtype=float32] 
        1D NumPy array of height values (meters)
    u_wind : numpy.ndarray[dtype=float32]
        1D NumPy array of environment U-wind components (m/s)
    v_wind : numpy.ndarray[dtype=float32]
        1D NumPy array of environment V-wind components (m/s)

    Returns
    -------
    float
        Storm Relative Helicity (m^2/s^2)
    """

@overload
def wind_shear(layer: nwsspc.sharp.calc.layer.HeightLayer, height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], u_wind: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], v_wind: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> WindComponents:
    """
    Computes the U and V components of the wind shear over a 
    layer given the vertical sounding arrays of height, u_wind,
    and v_wind.

    Parameters
    ----------
    layer : nwsspc.sharp.calc.layer.HeightLayer 
        HeightLayer for which to compute wind shear 
    height : numpy.ndarray[dtype=float32] 
        1D NumPy array of height values (meters)
    u_wind : numpy.ndarray[dtype=float32] 
        1D NumPy array of U-wind components (m/s)
    v_wind : numpy.ndarray[dtype=float32] 
        1D NumPy array of V-wind components (m/s)

    Returns
    -------
    nwsspc.sharp.calc.winds.WindComponents
        WindComponents of U and V wind shear components (m/s)
    """

@overload
def wind_shear(layer: nwsspc.sharp.calc.layer.PressureLayer, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], u_wind: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], v_wind: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> WindComponents:
    """
    Computes the U and V components of the wind shear over a 
    layer given the vertical sounding arrays of height, u_wind,
    and v_wind.

    Parameters
    ----------
    layer : nwsspc.sharp.calc.layer.PressureLayer 
        PressureLayer for which to compute wind shear 
    height : numpy.ndarray[dtype=float32] 
        1D NumPy array of height values (meters)
    u_wind : numpy.ndarray[dtype=float32] 
        1D NumPy array of U-wind components (m/s)
    v_wind : numpy.ndarray[dtype=float32] 
        1D NumPy array of V-wind components (m/s)

    Returns
    -------
    nwsspc.sharp.calc.winds.WindComponents
        WindComponents of U and V wind shear components (m/s)
    """

def mean_wind(layer: nwsspc.sharp.calc.layer.PressureLayer, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], u_wind: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], v_wind: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu')], weighted: bool = False) -> WindComponents:
    """
    Computes the mean wind over the given PressureLayer and input profile
    arrays of pressure, U-wind, and V-wind components. 

    Parameters
    ----------
    layer : nwsspc.sharp.calc.layer.PressureLayer 
        PressureLayer over which to compute mean
    pressure : numpy.ndarray[dtype=float32]
        1D NumPy array of pressure coordinate values (Pa)
    u_wind : numpy.ndarray[dtype=float32] 
        1D NumPy array of U-wind component values (m/s)
    v_wind : numpy.ndarray[dtype=float32] 
        1D NumPy array of V-wind component values (m/s)
    weighted : bool 
        Boolean flag to compute pressure-weighted mean wind (default: False)

    Returns
    -------
    nwsspc.sharp.calc.winds.WindComponents
        WindComponents of U anf V mean wind components (m/s)
    """

@overload
def max_wind(layer: nwsspc.sharp.calc.layer.PressureLayer, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], u_wind: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], v_wind: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> WindComponents:
    """
    Finds the maximum wind speed over a given nwsspc.sharp.calc.layer.PressureLayer,
    returning the vector components.

    Parameters
    ----------
    layer : nwsspc.sharp.calc.layer.PressureLayer 
        PressureLayer over which to compute mean
    pressure : numpy.ndarray[dtype=float32]
        1D NumPy array of pressure coordinate values (Pa)
    u_wind : numpy.ndarray[dtype=float32] 
        1D NumPy array of U-wind component values (m/s)
    v_wind : numpy.ndarray[dtype=float32] 
        1D NumPy array of V-wind component values (m/s)

    Returns
    -------
    nwsspc.sharp.calc.winds.WindComponents
    """

@overload
def max_wind(layer: nwsspc.sharp.calc.layer.HeightLayer, height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], u_wind: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], v_wind: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> WindComponents:
    """
    Finds the maximum wind speed over a given nwsspc.sharp.calc.layer.HeightLayer,
    returning the vector components.

    Parameters
    ----------
    layer : nwsspc.sharp.calc.layer.HeightLayer 
        PressureLayer over which to compute mean
    height : numpy.ndarray[dtype=float32]
        1D NumPy array of height coordinate values (m)
    u_wind : numpy.ndarray[dtype=float32] 
        1D NumPy array of U-wind component values (m/s)
    v_wind : numpy.ndarray[dtype=float32] 
        1D NumPy array of V-wind component values (m/s)

    Returns
    -------
    nwsspc.sharp.calc.winds.WindComponents
    """

@overload
def vector_magnitude(u_comp: float, v_comp: float) -> float:
    """
    Given the zonal (U) and meridional (V) wind components of a vector,
    compute and return the magnitude (m/s) of the vector.

    Parameters
    ----------
    u_comp : float
        U-wind component (m/s)
    v_comp : float
        V-wind component (m/s)

    Returns
    -------
    float
        Wind speed (m/s)
    """

@overload
def vector_magnitude(u_comp: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], v_comp: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')]:
    """
    Given the zonal (U) and meridional (V) components of a vector, 
    compute and return the magnitude (m/s) of the vector. 

    Parameters
    ----------
    u_comp : numpy.ndarray[dtype=float32] 
        1D NumPy array of U-wind component (m/s)
    v_comp : numpy.ndarray[dtype=float32] 
        1D NumPy array of V-wind component (m/s)

    Returns
    -------
    numpy.ndarray[dtype=float32]
        1D NumPy array of wind speed (m/s)
    """

@overload
def vector_angle(u_comp: float, v_comp: float) -> float:
    """
    Given the zonal (U) and meridional (V) components of a vector,
    compute and return the angle (from North) of the vector. 

    Parameters
    ----------
    u_comp : float 
        The U-wind component
    v_comp : float 
        The V-wind component

    Returns
    -------
    float
        Wind direction (degrees from North)
    """

@overload
def vector_angle(u_comp: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], v_comp: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')]:
    """
    Given the zonal (U) and meridional (V) components of a vector, 
    compute and return the angle (from North) of the vector. 

    Parameters
    ----------
    u_comp : numpy.ndarray[dtype=float32] 
        1D NumPy array of U-wind component (m/s)
    v_comp : numpy.ndarray[dtype=float32] 
        1D NumPy array of V-wind component (m/s)

    Returns
    -------
    numpy.ndarray[dtype=float32]
        1D NumPy array of wind direction (degrees from North)
    """

@overload
def u_component(wind_speed: float, wind_direction: float) -> float:
    """
    Computes the zonal (U) wind component from a wind vector.

    Parameters
    ----------
    wind_speed : float 
        The vector speed (m/s)
    wind_direction : float 
        The vector direction (degrees from North)

    Returns
    -------
    float
        The U-wind component (m/s)
    """

@overload
def u_component(wind_speed: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], wind_direction: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')]:
    """
    Computes the zonal (U) wind component from a wind vector.

    Parameters
    ----------
    wind_speed : numpy.ndarray[dtype=float32] 
        1D NumPy array of wind speeds (m/s)
    wind_direction : numpy.ndarray[dtype=float32] 
        1D NumPy array of wind direction (degrees)

    Returns
    -------
    numpy.ndarray[dtype=float32]
        1D NumPy array of U-wind component values (m/s)
    """

@overload
def v_component(wind_speed: float, wind_direction: float) -> float:
    """
    Computes the meridional (V) wind component from a wind vector.

    Parameters
    ----------
    wind_speed : float 
        Vector speed (m/s)
    wind_direction : float 
        Vector angle (degrees from North)

    Returns
    -------
    float
        The V-wind component
    """

@overload
def v_component(wind_speed: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], wind_direction: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')]:
    """
    Computes the meridional (V) wind component from a wind vector.

    Parameters
    ----------
    wind_speed : numpy.ndarray[dtype=float32] 
        1D NumPy array of wind speeds (m/s)
    wind_direction : numpy.ndarray[dtype=float32] 
        1D NumPy array of wind direction (degrees)

    Returns
    -------
    numpy.ndarray[dtype=float32]
        1D NumPy array of V-wind component values (m/s)
    """

@overload
def components_to_vector(u_comp: float, v_comp: float) -> WindVector:
    """
    Given the zonal (U) and meridional (V) components of a vector, 
    compute and return the wind speed (m/s) and direction 
    (degrees from North) as a WindVector type.

    Parameters
    ----------
    u_comp : float 
        The U-wind component (m/s)
    v_comp : float 
        The V-wind component (m/s)

    Returns
    -------
    nwsspc.sharp.calc.winds.WindVector
        WindVector containing wind speed (m/s) and direction (degrees from North)
    """

@overload
def components_to_vector(wind_comp: WindComponents) -> WindVector:
    """
    Given the components of a vector via a WindComponents object (m/s),
    compute and return the wind speed (m/s) and wind direction (degrees from North)
    as a WindVector type.

    Parameters
    ----------
    wind_comp : nwsspc.sharp.calc.winds.WindComponents 
        WindComponents (m/s)

    Returns
    -------
    nwsspc.sharp.calc.winds.WindVector
        WindVector containing wind speed (m/s) and direction (degrees from North)
    """

@overload
def components_to_vector(u_comp: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], v_comp: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> tuple[Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')], Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')]]:
    """
    Given 1D NumPy arrays of zonal (U) and meridional (V) wind components,
    compute the wind speed (m/s) and direction (degrees from North).

    Parameters
    ----------
    u_comp : numpy.ndarray[dtype=float32] 
        1D NumPy array of U-wind component values (m/s)
    v_comp : numpy.ndarray[dtype=float32]
        1D NumPy array of V-wind component values (m/s)

    Returns
    -------
    tuple[numpy.ndarray[dtype=float32], numpy.ndarray[dtype=float32]]
        wspd: 1D NumPy array of wind speeds (m/s)

        wdir: 1D NumPy array of wind directiond (degrees from North)
    """

@overload
def vector_to_components(wind_speed: float, wind_direction: float) -> WindComponents:
    """
    Given the wind speed (m/s) and wind direction (degrees from North),
    compute and return the zonal and meridional vector components as WindComponents.

    Parameters
    ----------
    wind_speed : float 
        The magnitude/speed of the vector (m/s)
    wind_direction : float 
        The direction of the vector (degrees from North)

    Returns
    -------
    nwsspc.sharp.calc.winds.WindComponents
        The U-wind and V-wind components (m/s) as WindComponents
    """

@overload
def vector_to_components(wind_vector: WindVector) -> WindComponents:
    """
    Given the wind speed (m/s) and direction (degrees from North),
    compute and return the zonal (U) and meridional (V) vector 
    components as WindComponents.

    Parameters
    ----------
    wind_vector : nwsspc.sharp.calc.winds.WindVector 
        WindVector containing wind wpeed (m/s) and direction (degrees from North)

    Returns
    -------
    nwsspc.sharp.calc.winds.WindComponents
        U and V wind components (m/s) in a WindComponents object
    """

@overload
def vector_to_components(wspd: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], wdir: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> tuple[Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')], Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')]]:
    """
    Given 1D NumPy arrays of the wind speed (m/s) and direction (degrees from North),
    compute and return 1D NumPy arays of the zonal (U) and meridional (V) vector 
    componenWindComponents.

    Parameters
    ----------
    wspd : numpy.ndarray[dtype=float32] 
        1D NumPy array of wind speeds (m/s)
    wdir : numpy.ndarray[dtype=float32]
        1D NumPy array of wind directions (degrees from North)

    Returns
    -------
    tuple[numpy.ndarray[dtype=float32], numpy.ndarray[dtype=float32]]
        uwin: 1D NumPy array of U-wind components

        vwin: 1D NumPy array of V-wind components
    """




import os
import pytest
import numpy as np
import pandas as pd

from nwsspc.sharp.calc import interp
from nwsspc.sharp.calc import layer
from nwsspc.sharp.calc import parcel
from nwsspc.sharp.calc import params
from nwsspc.sharp.calc import thermo
from nwsspc.sharp.calc import winds
from nwsspc.sharp.calc import constants


def load_parquet(filename):
    snd_df = pd.read_parquet(filename)
    snd_df = snd_df[snd_df["vwin"].notna()]
    snd_df = snd_df[snd_df["tmpc"].notna()]
    snd_df = snd_df[snd_df["relh"].notna()]
    snd_df = snd_df[snd_df["pres"] >= 50.0]

    pres = snd_df["pres"].to_numpy().astype('float32')*np.float32(100.0)
    hght = snd_df["hght"].to_numpy().astype('float32')
    tmpk = snd_df["tmpc"].to_numpy().astype('float32')+np.float32(273.15)
    dwpk = snd_df["dwpc"].to_numpy().astype('float32')+np.float32(273.15)
    relh = snd_df["relh"].to_numpy().astype('float32')
    wdir = snd_df["wdir"].to_numpy().astype('float32')
    wspd = snd_df["wspd"].to_numpy().astype('float32')
    uwin = snd_df["uwin"].to_numpy().astype('float32')
    vwin = snd_df["vwin"].to_numpy().astype('float32')

    # turn into height above ground level
    hght -= hght[0]

    # TO-DO - need a better interface to the API for doing this
    # uwin = np.empty(wspd.shape, dtype="float32")
    # vwin = np.empty(wspd.shape, dtype="float32")
    mixr = thermo.mixratio(pres, dwpk)
    vtmp = thermo.virtual_temperature(tmpk, mixr)
    theta = thermo.theta(pres, tmpk)
    thetae = thermo.thetae(
        pres,
        tmpk,
        dwpk
    )

    return {
        "pres": pres, "hght": hght,
        "tmpk": tmpk, "mixr": mixr,
        "relh": relh,
        "theta": theta,
        "thetae": thetae,
        "vtmp": vtmp, "dwpk": dwpk,
        "wdir": wdir, "wspd": wspd,
        "uwin": uwin, "vwin": vwin
    }


data_dir = os.path.join(
    os.path.dirname(__file__),
    "..",
    "..",
    "data",
    "test_snds"
)
filename = os.path.join(
    data_dir,
    "ddc.parquet"
)
snd_data = load_parquet(filename)


def test_effective_inflow_layer_wobus():

    lifter = parcel.lifter_wobus()
    mupcl = parcel.Parcel()
    eil = params.effective_inflow_layer(
        lifter,
        snd_data["pres"],
        snd_data["hght"],
        snd_data["tmpk"],
        snd_data["dwpk"],
        snd_data["vtmp"],
        mupcl=mupcl
    )

    assert (eil.bottom == pytest.approx(92043.0))
    assert (eil.top == pytest.approx(83384.0))
    assert (mupcl.cape == pytest.approx(3353.4, abs=5e-1))
    assert (mupcl.cinh == pytest.approx(-34.5707, abs=5e-4))


def test_effective_inflow_layer_cm1():
    mupcl = parcel.Parcel()
    lifter = parcel.lifter_cm1()
    lifter.ma_type = thermo.adiabat.pseudo_liq
    eil = params.effective_inflow_layer(
        lifter,
        snd_data["pres"],
        snd_data["hght"],
        snd_data["tmpk"],
        snd_data["dwpk"],
        snd_data["vtmp"],
        mupcl=mupcl
    )

    assert (eil.bottom == pytest.approx(92043.0))
    assert (eil.top == pytest.approx(83432.0))
    assert (mupcl.cape == pytest.approx(3107.6428, abs=5e-1))
    assert (mupcl.cinh == pytest.approx(-36.41, abs=5e-1))


def test_bunkers_motion_nonparcel():
    # Non parcel based Bunkers motion
    shr_lyr = layer.HeightLayer(0, 6000.0)
    mw_lyr = layer.HeightLayer(0, 6000)
    storm_mtn = params.storm_motion_bunkers(
        snd_data["pres"],
        snd_data["hght"],
        snd_data["uwin"],
        snd_data["vwin"],
        mw_lyr, shr_lyr,
    )

    assert (storm_mtn.u == pytest.approx(10.232333))
    assert (storm_mtn.v == pytest.approx(5.7385511))


def test_bunkers_motion():
    # parcel based Bunkers motion
    mupcl = parcel.Parcel()
    lifter = parcel.lifter_cm1()
    lifter.ma_type = thermo.adiabat.pseudo_liq
    eil = params.effective_inflow_layer(
        lifter,
        snd_data["pres"],
        snd_data["hght"],
        snd_data["tmpk"],
        snd_data["dwpk"],
        snd_data["vtmp"],
        mupcl=mupcl
    )

    storm_mtn = params.storm_motion_bunkers(
        snd_data["pres"],
        snd_data["hght"],
        snd_data["uwin"],
        snd_data["vwin"],
        eil, mupcl
    )

    assert (storm_mtn.u == pytest.approx(9.65811))
    assert (storm_mtn.v == pytest.approx(5.558156))


def test_corfidi_vectors():
    upshear, downshear = params.mcs_motion_corfidi(
        snd_data["pres"],
        snd_data["hght"],
        snd_data["uwin"],
        snd_data["vwin"]
    )

    assert (upshear.u == pytest.approx(12.7017, abs=1e-3))
    assert (upshear.v == pytest.approx(2.99329, abs=1e-3))
    assert (downshear.u == pytest.approx(23.2054, abs=1e-3))
    assert (downshear.v == pytest.approx(16.10519, abs=1e-3))


def test_effective_bulk_wind():
    lifter = parcel.lifter_cm1()
    lifter.ma_type = thermo.adiabat.pseudo_liq
    mupcl = parcel.Parcel()
    eil = params.effective_inflow_layer(
        lifter,
        snd_data["pres"],
        snd_data["hght"],
        snd_data["tmpk"],
        snd_data["dwpk"],
        snd_data["vtmp"],
        mupcl=mupcl
    )

    ebwd_cmp = params.effective_bulk_wind_difference(
        snd_data["pres"],
        snd_data["hght"],
        snd_data["uwin"],
        snd_data["vwin"],
        eil,
        mupcl.eql_pressure
    )

    ebwd = winds.vector_magnitude(ebwd_cmp.u, ebwd_cmp.v)
    assert (ebwd_cmp.u == pytest.approx(14.6, abs=1e-3))
    assert (ebwd_cmp.v == pytest.approx(13.321, abs=1e-3))
    assert (ebwd == pytest.approx(19.764, abs=1e-3))


def test_stp_scp_ship_dcp_lhp():
    lifter = parcel.lifter_cm1()
    lifter.ma_type = thermo.adiabat.pseudo_liq
    # get the mixed-layer parcel
    mix_lyr = layer.PressureLayer(
        snd_data["pres"][0], snd_data["pres"][0] - 10000.0)
    theta = thermo.theta(snd_data["pres"], snd_data["tmpk"])
    pcl = parcel.Parcel.mixed_layer_parcel(
        mix_lyr,
        snd_data["pres"],
        theta,
        snd_data["mixr"]
    )

    search_layer = layer.PressureLayer(
        snd_data["pres"][0],
        snd_data["pres"][0] - 40000.0
    )

    dcape_pcl = parcel.DowndraftParcel.min_thetae(
        search_layer,
        snd_data["pres"],
        snd_data["tmpk"],
        snd_data["dwpk"],
        snd_data["thetae"]
    )

    dpcl_t = dcape_pcl.lower_parcel(lifter, snd_data["pres"])
    dpcl_buoy = thermo.buoyancy(dpcl_t, snd_data["tmpk"])

    dcape, dcinh = dcape_pcl.cape_cinh(
        snd_data["pres"], snd_data["hght"], dpcl_buoy)

    # lift the parcel and get CAPE
    vtmpk = pcl.lift_parcel(lifter, snd_data["pres"])
    buoy = thermo.buoyancy(vtmpk, snd_data["vtmp"])
    cape, cinh = pcl.cape_cinh(snd_data["pres"], snd_data["hght"], buoy)

    # Get the effective inflow layer for effective SRH
    mupcl = parcel.Parcel()
    eil = params.effective_inflow_layer(
        lifter,
        snd_data["pres"],
        snd_data["hght"],
        snd_data["tmpk"],
        snd_data["dwpk"],
        snd_data["vtmp"],
        mupcl=mupcl
    )

    # Get the storm relative helicity for the effective inflow layer
    storm_mtn = params.storm_motion_bunkers(
        snd_data["pres"],
        snd_data["hght"],
        snd_data["uwin"],
        snd_data["vwin"],
        eil, mupcl
    )
    esrh = winds.helicity(
        eil,
        storm_mtn,
        snd_data["pres"],
        snd_data["uwin"],
        snd_data["vwin"]
    )

    ebwd_cmp = params.effective_bulk_wind_difference(
        snd_data["pres"],
        snd_data["hght"],
        snd_data["uwin"],
        snd_data["vwin"],
        eil,
        mupcl.eql_pressure
    )

    ebwd = winds.vector_magnitude(ebwd_cmp.u, ebwd_cmp.v)

    # Get the LCL height in meters AGL
    lcl_hght = interp.interp_pressure(
        pcl.lcl_pressure,
        snd_data["pres"],
        snd_data["hght"]
    ) - snd_data["hght"][0]

    stp = params.significant_tornado_parameter(
        pcl,
        lcl_hght,
        esrh,
        ebwd
    )
    assert (stp == pytest.approx(0.48329, abs=1e-4))

    scp = params.supercell_composite_parameter(mupcl.cape, esrh, ebwd)
    assert (scp == pytest.approx(7.9699, abs=1e-1))

    # get SHIP
    plyr = layer.PressureLayer(70000.0, 50000.0)
    hlyr = layer.HeightLayer(0.0, 6000.0)
    lr75 = thermo.lapse_rate(
        plyr, snd_data["pres"], snd_data["hght"], snd_data["tmpk"])
    t500 = interp.interp_pressure(50000.0, snd_data["pres"], snd_data["tmpk"])
    fzl = interp.find_first_height(
        constants.ZEROCNK, snd_data["hght"], snd_data["tmpk"])
    shr06 = winds.wind_shear(
        hlyr, snd_data["hght"], snd_data["uwin"], snd_data["vwin"])
    shr06 = winds.vector_magnitude(shr06.u, shr06.v)
    ship = params.significant_hail_parameter(mupcl, lr75, t500, fzl, shr06)
    assert (ship == pytest.approx(1.9521, abs=1e-3))

    hlyr_in_p = layer.height_layer_to_pressure(
        hlyr, snd_data["pres"], snd_data["hght"])
    mw06 = winds.mean_wind(
        hlyr_in_p, snd_data["pres"], snd_data["uwin"], snd_data["vwin"]
    )
    mw06 = winds.vector_magnitude(mw06.u, mw06.v)
    dcp = params.derecho_composite_parameter(dcape, mupcl.cape, shr06, mw06)
    assert (dcp == pytest.approx(7.63, abs=1e-1))

    hgz = params.hail_growth_layer(snd_data["pres"], snd_data["tmpk"])
    lhp = params.large_hail_parameter(
        mupcl,
        lr75,
        hgz,
        storm_mtn,
        snd_data["pres"],
        snd_data["hght"],
        snd_data["uwin"],
        snd_data["vwin"]
    )
    assert (lhp == pytest.approx(14.2394, abs=1e-2))


def test_ehi():
    pres = snd_data["pres"][0]
    tmpk = snd_data["tmpk"][0]
    dwpk = snd_data["dwpk"][0]

    pcl = parcel.Parcel.surface_parcel(pres, tmpk, dwpk)

    # Wobus Lifter
    lifter = parcel.lifter_wobus()
    vtmpk = pcl.lift_parcel(lifter, snd_data["pres"])
    buoy = thermo.buoyancy(vtmpk, snd_data["vtmp"])
    cape, cinh = pcl.cape_cinh(snd_data["pres"], snd_data["hght"], buoy)

    srh_lyr = layer.HeightLayer(0.0, 3000.0)
    mupcl = parcel.Parcel()
    eil = params.effective_inflow_layer(
        lifter,
        snd_data["pres"],
        snd_data["hght"],
        snd_data["tmpk"],
        snd_data["dwpk"],
        snd_data["vtmp"],
        mupcl=mupcl
    )

    # Get the storm relative helicity for the effective inflow layer
    storm_mtn = params.storm_motion_bunkers(
        snd_data["pres"],
        snd_data["hght"],
        snd_data["uwin"],
        snd_data["vwin"],
        eil, mupcl
    )
    srh = winds.helicity(
        srh_lyr,
        storm_mtn,
        snd_data["hght"],
        snd_data["uwin"],
        snd_data["vwin"]
    )

    ehi = params.energy_helicity_index(pcl.cape, srh)
    assert (ehi == pytest.approx(4.41228, abs=1e-5))


def test_convective_temperature():
    lifter = parcel.lifter_cm1()
    lifter.ma_type = thermo.adiabat.pseudo_liq
    cnvtv_tmpk = params.convective_temperature(
        lifter,
        snd_data["pres"],
        snd_data["hght"],
        snd_data["tmpk"],
        snd_data["vtmp"],
        snd_data["mixr"]
    )
    assert (cnvtv_tmpk == pytest.approx(304.65, abs=0.5))


def test_precipitable_water():
    plyr = layer.PressureLayer(snd_data["pres"][0], 40000.0)
    pwat = params.precipitable_water(plyr, snd_data["pres"], snd_data["mixr"])

    assert (pwat == pytest.approx(21.11469))


def test_hgz():
    hgz = params.hail_growth_layer(snd_data["pres"], snd_data["tmpk"])
    assert (hgz.bottom == pytest.approx(51430))
    assert (hgz.top == pytest.approx(35816))


def test_dgz():
    dgz = params.dendritic_layer(snd_data["pres"], snd_data["tmpk"])
    assert (dgz.bottom == pytest.approx(49598))
    assert (dgz.top == pytest.approx(45961))


def test_fwwi():
    fwwi = params.fosberg_fire_index(
        308,
        0.00001,
        13.5
    )
    assert (fwwi == 100.0)

    tmpk = np.array([308, 308, 308], dtype='float32')
    relh = np.array([0.00001, 0.00001, 0.00001], dtype='float32')
    wspd = np.array([13.5, 13.5, 13.5], dtype='float32')

    fwwi = params.fosberg_fire_index(tmpk, relh, wspd)
    assert (fwwi == np.array([100.0, 100.0, 100.0], dtype='float32')).all()


def test_pft():
    lifter = parcel.lifter_cm1()
    lifter.ma_type = thermo.adiabat.pseudo_liq
    mix_layer = layer.PressureLayer(
        snd_data["pres"][0], snd_data["pres"][0] - 10000.0)
    pft = params.pyrocumulonimbus_firepower_threshold(
        lifter,
        mix_layer,
        snd_data["pres"],
        snd_data["hght"],
        snd_data["tmpk"],
        snd_data["mixr"],
        snd_data["vtmp"],
        snd_data["uwin"],
        snd_data["vwin"],
        snd_data["theta"]
    )
    assert (pft == pytest.approx(153856081920.0, abs=1e6))



import enum
from typing import Annotated, overload

import numpy
from numpy.typing import NDArray

import nwsspc.sharp.calc.layer
import nwsspc.sharp.calc.thermo


class lifter_wobus:
    """
    A functor that calls the Wobus Wetlift for computation of moist adiabats.

    This is used to wrap the Wobus Wetlift function for parcel lifting 
    routines. Functors -- classes with their operator() overloaded -- 
    are used so that functions can be passed to templates in a way that 
    the compiler can still optimize, rather than using function pointers
    or lambdas. 

    Specifically, this functor is designed to be passed as a template 
    argument to nwsspc.sharp.calc.parcel.Parcel.lift_parcel, so that 
    the method of computing moist adiabats can be changed without 
    changing the overall parcel lifting code. The reason this is awesome 
    is that the compiler can still optimize and inline this code, while 
    the user can configure the parcel lifting algorithm to their 
    specifications.
    """

    def __init__(self) -> None: ...

    lift_from_lcl: bool = ...
    """
    A static flag that helps the parcel lifting functions know where to lift from.
    """

    @property
    def converge(self) -> float:
        """The iterative convergence criteria (K)"""

    @converge.setter
    def converge(self, arg: float, /) -> None: ...

    def setup(self, lcl_pres: float, lcl_tmpk: float) -> None:
        """
        Some parcel lifters require setup in order to handle adiabatic ascent
        when tracking the water vapor, liquid, and ice mixing ratios. The Wobus 
        lifter does not require this, however, so this function does nothing.
        """

    def __call__(self, pres: float, tmpk: float, new_pres: float) -> float:
        """
        Overloads the call operator in order to facilitate parcel lifting. 

        Parameters
        ----------
        pres : float 
            Parcel pressure (Pa)
        tmpk : float 
            Parcel temperature (K)
        new_pres : float 
            Final level of parcel after lift (Pa)

        Returns
        -------
        float
            The temperature of the lifted parcel (K)
        """

    def parcel_virtual_temperature(self, pres: float, tmpk: float) -> float:
        """
        Computes the virtual temperature of the parcel (after saturation).

        Parameters
        ----------
        pres : float 
            Parcel pressure (Pa)
        tmpk : float 
            Parcel temperature (K)

        Returns
        -------
        float
            The virtual temperature of the parcel (K)
        """

class lifter_cm1:
    """
    Use the CM1 moist lift calculations for adiabatic and pseudoadiabatic
    parcel ascent.
    """

    def __init__(self) -> None: ...

    lift_from_lcl: bool = ...
    """
    A static flag that helps the parcel lifting functions know where to lift from.
    The lifter_cm1 lifts from the last lifted level, rather than the LCL, because
    it is an iterative solver. This results in major performance improvements while 
    maintaining accuracy.
    """

    @property
    def ma_type(self) -> nwsspc.sharp.calc.thermo.adiabat:
        """
        The type of moist adiabat to use, defined by nwsspc.sharp.calc.thermo.adiabat
        """

    @ma_type.setter
    def ma_type(self, arg: nwsspc.sharp.calc.thermo.adiabat, /) -> None: ...

    @property
    def pressure_incr(self) -> float:
        """The pressure increment (Pa) to use for the iterative solver."""

    @pressure_incr.setter
    def pressure_incr(self, arg: float, /) -> None: ...

    @property
    def converge(self) -> float:
        """The iterative convergence criteria (K)"""

    @converge.setter
    def converge(self, arg: float, /) -> None: ...

    def setup(self, lcl_pres: float, lcl_tmpk: float) -> None:
        """
        This function sets the total water mixing ratio for 
        adiabatic parcel ascent, and zeroes out the vapor,
        liquid, and ice mixing ratio from previous parcel 
        ascents. 

        Parameters
        ----------
        lcl_pres : float 
            The LCL pressure (Pa)
        lcl_tmpk : float 
            The LCL temperature (K)

        Returns
        -------
        None
        """

    def __call__(self, pres: float, tmpk: float, new_pres: float) -> float:
        """
        Lifts a parcel moist adiabatically/pseudoadiabatically using
        nwsspc.sharp.calc.thermo.moist_adiabat_cm1.

        Parameters
        ----------
        pres : float 
            Parcel pressure (Pa)
        tmpk : float 
            Parcel temperature (K)
        new_pres : float 
            Final level of parcel after lift (Pa)

        Returns
        -------
        float
            The temperature of the lifted parcel (K)
        """

    def parcel_virtual_temperature(self, pres: float, tmpk: float) -> float:
        """
        Computes the virtual temperature of the parcel (after saturation).

        Parameters
        ----------
        pres : float 
            Parcel pressure (Pa)
        tmpk : float 
            Parcel temperature (K)

        Returns
        -------
        float
            The virtual temperature of the parcel (K)
        """

class LPL(enum.Enum):
    SFC = 1
    """A Surface Based Parcel"""

    FCST = 2
    """A Forecast Surface Parcel"""

    MU = 3
    """A Most Unstable Parcel"""

    ML = 4
    """A Mixed-Layer Parcel"""

    USR = 5
    """A User Defined Parcel"""

class Parcel:
    """
    Contains information about a Parcel's starting level and
    thermodynamic attributes, as well as derived computations,
    methods for constructing a parcel, and parcel ascent routines.
    """

    @overload
    def __init__(self) -> None: ...

    @overload
    def __init__(self, pressure: float, temperature: float, dewpoint: float, lpl: LPL) -> None:
        """
        Constructor for a Parcel

        Parameters
        ----------
        pressure : float 
            Parcel initial pressure (Pa)
        temperature : float 
            Parcel initial temperature (K)
        dewpoint : float 
            Parcel initial dewpoint (K)
        lpl : nwsspc.sharp.calc.parcel.LPL 
            Parcel Lifted Parcel Level (LPL) definition
        """

    @property
    def pres(self) -> float:
        """Parcel starting pressure (Pa)"""

    @pres.setter
    def pres(self, arg: float, /) -> None: ...

    @property
    def tmpk(self) -> float:
        """Parcel starting temperature (K)"""

    @tmpk.setter
    def tmpk(self, arg: float, /) -> None: ...

    @property
    def dwpk(self) -> float:
        """Parcel starting dewpoint (K)"""

    @dwpk.setter
    def dwpk(self, arg: float, /) -> None: ...

    @property
    def lcl_pressure(self) -> float:
        """Pressure at the Lifted Condensation Level (Pa)"""

    @lcl_pressure.setter
    def lcl_pressure(self, arg: float, /) -> None: ...

    @property
    def lfc_pressure(self) -> float:
        """Pressure at the Level of Free Convection (Pa)"""

    @lfc_pressure.setter
    def lfc_pressure(self, arg: float, /) -> None: ...

    @property
    def eql_pressure(self) -> float:
        """Pressure at the parcel Equilibrium Level"""

    @eql_pressure.setter
    def eql_pressure(self, arg: float, /) -> None: ...

    @property
    def mpl_pressure(self) -> float:
        """Pressure at the Maximum Parcel Level"""

    @mpl_pressure.setter
    def mpl_pressure(self, arg: float, /) -> None: ...

    @property
    def cape(self) -> float:
        """
        Parcel Convective Available Potential Energy (J/kg) between the LFC and EL
        """

    @cape.setter
    def cape(self, arg: float, /) -> None: ...

    @property
    def cinh(self) -> float:
        """Parcel Convective Inhibition (J/kg) between the LFC and EL"""

    @cinh.setter
    def cinh(self, arg: float, /) -> None: ...

    @overload
    def lift_parcel(self, lifter: lifter_wobus, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')]:
        """
        Lifts a Parcel dry adiabatically from its LPL to its LCL dry
        adiabatically, and then moist adiabatically from the LCL to 
        the top of the profile. The moist adiabat used is determined
        by the type of lifting functor passed to the function (i.e.
        lifter_wobus or lifter_cm1).

        Parameters
        ----------
        lifter : nwsspc.sharp.calc.parcel.lifter_wobus 
            An instantiated lifter_wobus functor
        pressure : numpy.ndarray[dtype=float32] 
            1D NumPy array of Pressure levels for lifting (Pa)

        Returns
        -------
        numpy.ndarray[dtype=float32]
            A 1D NumPy array of parcel virtual temperature values (K)
        """

    @overload
    def lift_parcel(self, lifter: lifter_cm1, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')]:
        """
        Lifts a Parcel dry adiabatically from its LPL to its LCL dry
        adiabatically, and then moist adiabatically from the LCL to 
        the top of the profile. The moist adiabat used is determined
        by the type of lifting functor passed to the function (i.e.
        lifter_wobus or lifter_cm1).

        Parameters
        ----------
        lifter : nwsspc.sharp.calc.parcel.lifter_cm1 
            An instantiated lifter_cm1 functor
        pressure : numpy.ndarray[dtype=float32] 
            1D NumPy array of Pressure levels for lifting (Pa)

        Returns
        -------
        numpy.ndarray[dtype=float32]
            A 1D NumPy array of parcel virtual temperature values (K)
        """

    def find_lfc_el(self, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], buoyancy: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> tuple[float, float]:
        """
        Searches the buoyancy array for the LFC and EL combination that results
        in the maximum amount of CAPE in the given profile. The buoyancy array 
        is typically computed by calling nwsspc.sharp.calc.parcel.Parcel.lift_parcel. 
        Once the LFC and EL are found, the value are set in 
        nwsspc.sharp.calc.parcel.Parcel.lfc_pres and 
        nwsspc.sharp.calc.parcel.Parcel.eql_pres via the provided parcel.

        The value of eql_pres is MISSING if there is no qualifying level 
        found within the data bounds (e.g. incomplete data, or EL above 
        the available data). Any calls to nwsspc.sharp.calc.parcel.Parcel.cape_cinh 
        will still compute CAPE without the presence of an EL, using the best-available
        data. 

        Parameters
        ----------
        pressure : numpy.ndarray[dtype=float32] 
            1D NumPy array of pressure values (Pa)
        height : numpy.ndarray[dtype=float32]
            1D NumPy array of height values (meters)
        buoyancy : numpy.ndarray[dtype=float32] 
            1D NumPy array of buoyancy values (m/s^2)

        Returns
        -------
        tuple[float, float]
            (LFC_PRES, EL_PRES)
        """

    def maximum_parcel_level(self, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], buoyancy: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> float:
        """
        Find the pressure of the Maximum Parcel Level (MPL).

        The Maximum Parcel Level (MPL) is the level a parcel woud reach 
        if it expended all of its integrated positive buoyancy past the 
        Equilibrium Level. It is found by integrating negatively buoyant 
        area above the Equilibrium Level until the integrated negative 
        buoyancy is equal in magnitude to the Convective Available 
        Potential Energy between the Level of Free Convection and the 
        Equilibrium Level. 

        For valid calculations, nwsspc.sharp.calc.parcel.Parcel.cape_cinh 
        must be called first, or nwsspc.sharp.calc.parcel.Parcel.cape and 
        nwsspc.sharp.calc.parcel.Parcel.eql_pressure must be set. 

        A values of nwsspc.sharp.calc.constants.MISSING is returned if:
          * CAPE is 0 
          * nwsspc.sharp.calc.parce.Parcel.eql_pressure is MISSING
          * No valid MPL candidate is found within the profile
            In this scenario, it likely exceeds the top of the available data

        In addition to being returned, the result is stored inside of 
        nwsspc.sharp.calc.parcel.Parcel.mpl_pressure.

        Parameters
        ----------
        pres : numpy.ndarray[dtype=float32] 
            1D NumPy array of pressure values (Pa)
        hght : numpy.ndarray[dtype=float32] 
            1D NumPy array of height values (meters)
        buoyancy : numpy.ndarray[dtype=float32] 
            1D NumPy array of buoyancy values (m/s^2)

        Returns 
        -------
        float 
            The pressure of the Maximum Parcel Level (Pa)
        """

    def cape_cinh(self, pres: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], hght: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], buoy: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> tuple[float, float]:
        """
        Assuming that nwsspc.sharp.calc.parcel.Parcel.lift_parcel has 
        been called, cape_cinh will integrate the area between the LFC 
        and EL to compute CAPE, and integrate the area between the LPL 
        and LCL to compute CINH.

        If eql_pressure is MISSING, but lfc_pressure is defined, the 
        routine will compute CAPE with the available data despite the 
        lack of a defined equilibrium level. This is useful for 
        incomplete profile data, or pressure-level data where the EL 
        is above the top pressure value.

        The results are stored in nwsspc.sharp.calc.parcel.Parcel.cape 
        and nwsspc.sharp.calc.parcel.Parcel.cinh via the provided parcel.

        Parameters
        ----------
        pres : numpy.ndarray[dtype=float32] 
            1D NumPy array of pressure values (Pa)
        hght : numpy.ndarray[dtype=float32] 
            1D NumPy array of height values (meters)
        buoyancy : numpy.ndarray[dtype=float32] 
            1D NumPy array of buoyancy values (m/s^2)

        Returns
        -------
        tuple[float, float]
            (CAPE, CINH)
        """

    def lifted_index(self, pres_lev: float, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], virtual_temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], parcel_virtual_temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> float:
        """
        Using the parcel and environment virtual temperature, compute the lifted index 
        at the given level. The pressure level typically used is 500 hPa. The lifted index 
        is the difference between the environment and parcel virtual temperatures at the 
        requested level. 

        Parameters
        ----------
        pres_lev : float 
            The pressure level (Pa) to compute the lifted index at
        pressure : numpy.ndarray[dtype=float32]
            1D NumPy array of pressure values (Pa)
        virtual_temperature : numpy.ndarray[dtype=float32]
            1D NumPy array of environment virtual temperature values (K)
        parcel_virtual_temperature : numpy.ndarray[dtype=float32]
            1D NumPy array of parcel virtual temperature values (K)

        Returns 
        -------
        float 
            The lifted index
        """

    @staticmethod
    def surface_parcel(pressure: float, temperature: float, dewpoint: float) -> Parcel:
        """
        Given input values of surface pressure, temperature, and dewpoint 
        temperature, construct and return a Surface Based Parcel.

        Parameters
        ----------
        pressure : float 
            Surface pressure (Pa)
        temperature : float 
            Surface temperature (K)
        dewpoint : float 
            Surface dewpoint (K)

        Returns
        -------
        nwsspc.sharp.calc.parcel.Parcel
            Parcel with surface values
        """

    @overload
    @staticmethod
    def mixed_layer_parcel(mix_layer: nwsspc.sharp.calc.layer.PressureLayer, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], potential_temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], mixing_ratio: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Parcel:
        """
        Given input arrays of pressure, potential temperature, 
        and water vapor mixing ratio, as well as a defined PressureLayer, 
        compute and return a mixed-layer Parcel.

        Parameters
        ----------
        mix_layer : nwsspc.sharp.calc.layer.PressureLayer 
            PressureLayer over which to compute a mixed-layer parcel 
        pressure : numpy.ndarray[dtype=float32] 
            1D NumPy array of profile pressure values (Pa)
        potential_temperature : numpy.ndarray[dtype=float32] 
            1D NumPy array of profile potential temperature (K)
        mixing_ratio : numpy.ndarray[dtype=float32] 
            1D NumPy array of profile water vapor mixing ratio (unitless)

        Returns
        -------
        nwsspc.sharp.calc.parcel.Parcel
            Parcel with mixed layer values
        """

    @overload
    @staticmethod
    def mixed_layer_parcel(mix_layer: nwsspc.sharp.calc.layer.HeightLayer, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], potential_temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], mixing_ratio: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Parcel:
        """
        Given input arrays of pressure, potential temperature, and water 
        vapor mixing ratio, as well as a defined PressureLayer, compute 
        and return a mixed-layer Parcel.

        Parameters
        ----------
        mix_layer : nwsspc.sharp.calc.layer.HeightLayer 
            HeightLayer over which to compute a mixed-layer parcel 
        pressure : numpy.ndarray[dtype=float32] 
            1D NumPy array of profile pressure values (Pa)
        height : numpy.ndarray[dtype=float32] 
            1D NumPy array of profile height values (meters)
        potential_temperature : numpy.ndarray[dtype=float32] 
            1D NumPy array of profile potential temperature (K)
        mixing_ratio : numpy.ndarray[dtype=float32] 
            1D NumPy array of profile water vapor mixing ratio (unitless)

        Returns
        -------
        nwsspc.sharp.calc.parcel.Parcel
            Parcel with mixed layer values
        """

    @overload
    @staticmethod
    def most_unstable_parcel(layer: nwsspc.sharp.calc.layer.PressureLayer, lifter: lifter_cm1, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], virtual_temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], dewpoint: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Parcel:
        """
        Given input arrays of pressure, height, temperature, virtual temperature,
        and dewpoint temperature, as well as a defined PressureLayer/HeightLayer and 
        parcel lifter (lifter_wobus or lifter_cm1), find and return the most unstable parcel. 

        Parameters
        ----------
        layer : nwsspc.sharp.calc.layer.PressureLayer 
            PressureLayer for which to search for the Most Unstable Parcel
        lifter : nwsspc.sharp.calc.parcel.lifter_cm1 
            Parcel lifting routine to use for moist ascent
        pressure : numpy.ndarray[dtype=float32] 
            1D NumPy array of profile pressure values (Pa)
        height : numpy.ndarray[dtype=float32] 
            1D NumPy array of profile height values (meters)
        temperature : numpy.ndarray[dtype=float32] 
            1D NumPy array of profile temperature values (K)
        virtual_temperature : numpy.ndarray[dtype=float32] 
            1D NumPy array of profile virtual temperature values (K)
        dewpoint : numpy.ndarray[dtype=float32] 
            1D NumPy array of profile dewpoint values (K)

        Returns
        -------
        nwsspc.sharp.calc.parcel.Parcel
            Parcel with most-unstable values
        """

    @overload
    @staticmethod
    def most_unstable_parcel(layer: nwsspc.sharp.calc.layer.HeightLayer, lifter: lifter_cm1, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], virtual_temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], dewpoint: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Parcel:
        """
        Given input arrays of pressure, height, temperature, virtual temperature,
        and dewpoint temperature, as well as a defined PressureLayer/HeightLayer and 
        parcel lifter (lifter_wobus or lifter_cm1), find and return the most unstable parcel. 

        Parameters
        ----------
        layer : nwsspc.sharp.calc.layer.HeightLayer 
            HeightLayer for which to search for the Most Unstable Parcel
        lifter : nwsspc.sharp.calc.parcel.lifter_cm1 
            Parcel lifting routine to use for moist ascent
        pressure : numpy.ndarray[dtype=float32] 
            1D NumPy array of profile pressure values (Pa)
        height : numpy.ndarray[dtype=float32] 
            1D NumPy array of profile height values (meters)
        temperature : numpy.ndarray[dtype=float32] 
            1D NumPy array of profile temperature values (K)
        virtual_temperature : numpy.ndarray[dtype=float32] 
            1D NumPy array of profile virtual temperature values (K)
        dewpoint : numpy.ndarray[dtype=float32] 
            1D NumPy array of profile dewpoint values (K)

        Returns
        -------
        nwsspc.sharp.calc.parcel.Parcel
            Parcel with most-unstable values
        """

    @overload
    @staticmethod
    def most_unstable_parcel(layer: nwsspc.sharp.calc.layer.PressureLayer, lifter: lifter_wobus, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], virtual_temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], dewpoint: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Parcel:
        """
        Given input arrays of pressure, height, temperature, virtual temperature,
        and dewpoint temperature, as well as a defined PressureLayer/HeightLayer and 
        parcel lifter (lifter_wobus or lifter_cm1), find and return the most unstable parcel. 

        Parameters
        ----------
        layer : nwsspc.sharp.calc.layer.PressureLayer 
            PressureLayer for which to search for the Most Unstable Parcel
        lifter : nwsspc.sharp.calc.parcel.lifter_wobus 
            Parcel lifting routine to use for moist ascent
        pressure : numpy.ndarray[dtype=float32] 
            1D NumPy array of profile pressure values (Pa)
        height : numpy.ndarray[dtype=float32] 
            1D NumPy array of profile height values (meters)
        temperature : numpy.ndarray[dtype=float32] 
            1D NumPy array of profile temperature values (K)
        virtual_temperature : numpy.ndarray[dtype=float32] 
            1D NumPy array of profile virtual temperature values (K)
        dewpoint : numpy.ndarray[dtype=float32] 
            1D NumPy array of profile dewpoint values (K)

        Returns
        -------
        nwsspc.sharp.calc.parcel.Parcel
        """

    @overload
    @staticmethod
    def most_unstable_parcel(layer: nwsspc.sharp.calc.layer.HeightLayer, lifter: lifter_wobus, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], virtual_temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], dewpoint: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Parcel:
        """
        Given input arrays of pressure, height, temperature, virtual temperature,
        and dewpoint temperature, as well as a defined PressureLayer/HeightLayer and 
        parcel lifter (lifter_wobus or lifter_cm1), find and return the most unstable parcel. 

        Parameters
        ----------
        layer : nwsspc.sharp.calc.layer.HeightLayer 
            HeightLayer for which to search for the Most Unstable Parcel
        lifter : nwsspc.sharp.calc.parcel.lifter_wobus 
            Parcel lifting routine to use for moist ascent
        pressure : numpy.ndarray[dtype=float32] 
            1D NumPy array of profile pressure values (Pa)
        height : numpy.ndarray[dtype=float32] 
            1D NumPy array of profile height values (meters)
        temperature : numpy.ndarray[dtype=float32] 
            1D NumPy array of profile temperature values (K)
        virtual_temperature : numpy.ndarray[dtype=float32] 
            1D NumPy array of profile virtual temperature values (K)
        dewpoint : numpy.ndarray[dtype=float32] 
            1D NumPy array of profile dewpoint values (K)

        Returns
        -------
        nwsspc.sharp.calc.parcel.Parcel
            Parcel with most-unstable values
        """

class DowndraftParcel:
    """
    Contains information about a DowndraftParcel's starting level and
    thermodynamic attributes, as well as derived computations,
    methods for constructing a parcel, and parcel descent routines.
    """

    @overload
    def __init__(self) -> None: ...

    @overload
    def __init__(self, pressure: float, temperature: float, dewpoint: float) -> None:
        """
        Constructor for a DowndraftParcel

        Parameters
        ----------
        pressure : float 
            DowndraftParcel initial pressure (Pa)
        temperature : float 
            DowndraftParcel initial temperature (K)
        dewpoint : float 
            DowndraftParcel initial dewpoint (K)
        """

    @property
    def pres(self) -> float:
        """DowndraftParcel starting pressure (Pa)"""

    @pres.setter
    def pres(self, arg: float, /) -> None: ...

    @property
    def tmpk(self) -> float:
        """DowndraftParcel starting temperature (K)"""

    @tmpk.setter
    def tmpk(self, arg: float, /) -> None: ...

    @property
    def dwpk(self) -> float:
        """DowndraftParcel starting dewpoint (K)"""

    @dwpk.setter
    def dwpk(self, arg: float, /) -> None: ...

    @property
    def cape(self) -> float:
        """DowndraftParcel Convective Available Potential Energy (J/kg)"""

    @cape.setter
    def cape(self, arg: float, /) -> None: ...

    @property
    def cinh(self) -> float:
        """DowndraftParcel Convective Inhibition (J/kg)"""

    @cinh.setter
    def cinh(self, arg: float, /) -> None: ...

    @overload
    def lower_parcel(self, lifter: lifter_wobus, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')]:
        """
        Lowers a saturated nwsspc.sharp.calc.parcel.DowndraftParcel moist 
        adiabatically from its LPL to the surface. The moist adiabat used 
        is determined by the type of lifting functor passed to the function 
        (i.e. lifter_wobus or lifter_cm1).

        Unlike nwsspc.sharp.calc.parcel.Parcel.lift_parcel, the virtual 
        temperature correction is not used for downdraft parcels.

        Parameters
        ----------
        lifter : nwsspc.sharp.calc.parcel.lifter_wobus 
            An instantiated lifter_wobus functor
        pressure : numpy.ndarray[dtype=float32] 
            1D NumPy array of Pressure levels for lifting (Pa)

        Returns
        -------
        numpy.ndarray[dtype=float32]
            A 1D NumPy array of parcel temperature values (K)
        """

    @overload
    def lower_parcel(self, lifter: lifter_cm1, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')]:
        """
        Lowers a saturated nwsspc.sharp.calc.parcel.DowndraftParcel moist 
        adiabatically from its LPL to the surface. The moist adiabat used 
        is determined by the type of lifting functor passed to the function 
        (i.e. lifter_wobus or lifter_cm1).

        Unlike nwsspc.sharp.calc.parcel.Parcel.lift_parcel, the virtual 
        temperature correction is not used for downdraft parcels.

        Parameters
        ----------
        lifter : nwsspc.sharp.calc.parcel.lifter_cm1
            An instantiated lifter_cm1 functor
        pressure : numpy.ndarray[dtype=float32] 
            1D NumPy array of Pressure levels for lifting (Pa)

        Returns
        -------
        numpy.ndarray[dtype=float32]
            A 1D NumPy array of parcel temperature values (K)
        """

    def cape_cinh(self, pres: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], hght: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], buoy: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> tuple[float, float]:
        """
        Assuming that nwsspc.sharp.calc.parcel.DowndraftParcel.lower_parcel 
        has been called, cape_cinh will integrate the area between the LPL
        and the surface to compute downdraft CAPE and downdraft CINH.

        The results are stored in nwsspc.sharp.calc.parcel.DowndraftParcel.cape 
        and nwsspc.sharp.calc.parcel.DowndraftParcel.cinh via the provided parcel.

        Parameters
        ----------
        pres : numpy.ndarray[dtype=float32] 
            1D NumPy array of pressure values (Pa)
        hght : numpy.ndarray[dtype=float32] 
            1D NumPy array of height values (meters)
        buoyancy : numpy.ndarray[dtype=float32] 
            1D NumPy array of buoyancy values (m/s^2)

        Returns
        -------
        tuple[float, float]
            (DCAPE, DCINH)
        """

    @staticmethod
    def min_thetae(search_layer: nwsspc.sharp.calc.layer.PressureLayer, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], dewpoint: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], thetae: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], mean_depth: float = 10000.0) -> DowndraftParcel:
        """
        Define a downdraft parcel. 

        Defines a downdraft parcel within a given search layer. 
        The downdraft parcel is defined as the minimum layer-mean 
        equivalent potential temperature (Theta-E) within the 
        search layer. Typical values are to search within the lowest
        400 hPa of the profile, and a mean depth of 100 hPa. 

        Parameters 
        ----------
        search_layer : nwsspc.sharp.calc.layer.PressureLayer 
            The layer over which to search for the downdraft parcel 
        pressure : numpy.ndarray[dtype=float32]
            1D NumPy array of pressure (Pa)
        temperature : numpy.ndarray[dtype=float32]
            1D NumPy array of temperature (K)
        dewpoint : numpy.ndarray[dtype=float32]
            1D NumPy array of dewpoint (K)
        thetae : numpy.ndarray[dtype=float32]
            1D NumPy array of thetae (K)
        mean_depth : float
            The layer depth for calculating mean thetae.

        Returns 
        -------
        nwsspc.sharp.calc.parcel.DowndraftParcel 
            Downdraft Parcel
        """



import enum
from typing import Annotated, overload

import numpy
from numpy.typing import NDArray

import nwsspc.sharp.calc.layer
import nwsspc.sharp.calc.parcel


class adiabat(enum.Enum):
    pseudo_liq = 1
    """Use a pseudoadiabatic parcel with liquid only processes"""

    adiab_liq = 2
    """Use an adiabatic parcel with liquid only processes"""

    pseudo_ice = 3
    """Use a pseudoadiabatic parcel with liquid and ice processes"""

    adiab_ice = 4
    """Use an adiabatic parcel with liquid and ice processes"""

@overload
def wobf(temperature: float) -> float:
    """
    Computes the difference between the wet-bulb potential temperatures
    (Kelvin) for saturated and dry air, given the temperature (Kelvin).

    The Wobus Function (wobf) is defined as the difference between the 
    wet-bulb potential temperature for saturated air (WBPTS) and the 
    wet-bulb potential temperature for dry air (WBPTD) given an iput
    air temperature. This function is used in the computation of moist 
    adiabatic ascent as part of the Wobus parcel lifter (lifter_wobus).

    WOBF(T) := WBPTS - WBPTD

    Although WBPTS and WBPTD are functions of both pressure and 
    temperature, it is assumed their difference is a function of 
    temperature only. The difference is also proportional to the 
    heat imparted to a parcel of air.

    This function uses a polynomial approximation to the wobus function, 
    fitted to value in Table 78 of PP.319-322 of the Smithsonian Meteorological
    Table by Rolan List (6th Revised Edition). Herman Wobus, a mathematician 
    for the Navy Weather Research Facility in Norfolk, VA computed these 
    coefficients a very long time ago, as he was retired as of the time of 
    the documentation found on this routine written in 1981.

    It was shown by Robert Davies-Jones (2007) that the Wobus function has
    a slight dependence on pressure, which results in errors of up to 1.2
    Kelvin in the temperature of a lifted parcel. 

    Parameters
    ----------
    temperature : float 
        The air temperature (Kelvin)

    Returns
    -------
    float
        The Wobus function temperature (Kelvin)
    """

@overload
def wobf(tmpk_arr: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')]:
    """
    Computes the difference between the wet-bulb potential temperatures
    (Kelvin) for saturated and dry air, given the temperature (Kelvin).

    The Wobus Function (wobf) is defined as the difference between the 
    wet-bulb potential temperature for saturated air (WBPTS) and the 
    wet-bulb potential temperature for dry air (WBPTD) given an iput
    air temperature. This function is used in the computation of moist 
    adiabatic ascent as part of the Wobus parcel lifter (lifter_wobus).

    WOBF(T) := WBPTS - WBPTD

    Although WBPTS and WBPTD are functions of both pressure and 
    temperature, it is assumed their difference is a function of 
    temperature only. The difference is also proportional to the 
    heat imparted to a parcel of air.

    This function uses a polynomial approximation to the wobus function, 
    fitted to value in Table 78 of PP.319-322 of the Smithsonian Meteorological
    Table by Rolan List (6th Revised Edition). Herman Wobus, a mathematician 
    for the Navy Weather Research Facility in Norfolk, VA computed these 
    coefficients a very long time ago, as he was retired as of the time of 
    the documentation found on this routine written in 1981.

    It was shown by Robert Davies-Jones (2007) that the Wobus function has
    a slight dependence on pressure, which results in errors of up to 1.2
    Kelvin in the temperature of a lifted parcel. 

    Parameters
    ----------
    tmpk_arr : numpy.ndarray[dtype=float32] 
        The 1D array of air temperatures (Kelvin)

    Returns
    -------
    numpy.ndarray[dtype=float32]
        The 1D array of Wobus function temperatures (Kelvin)
    """

@overload
def lcl_temperature(temperature: float, dewpoint: float) -> float:
    """
    Compute the Lifted Condensation Level (LCL) temperature (Kelvin) 
    given an air temperature (Kelvin) and dewpoint temperature (Kelvin).

    The LCL temperature is computed as in Bolton (1980) eq 15, and is
    considered to be within a 10th of a degree of the more exact 
    iterative formula.

    Parameters
    ----------
    temperature : float 
        The air temperature (Kelvin)
    dewpoint : float 
        The dewpoint temperature (Kelvin)

    Returns
    -------
    float
        The LCL temperature (Kelvin)
    """

@overload
def lcl_temperature(tmpk_arr: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], dwpk_arr: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')]:
    """
    Compute the Lifted Condensation Level (LCL) temperature (Kelvin) 
    given an array of air temperatures (Kelvin) and dewpoint 
    temperatures (Kelvin).

    The LCL temperatures are computed as in Bolton (1980) eq 15, and is
    considered to be within a 10th of a degree of the more exact 
    iterative formula.

    Parameters
    ----------
    tmpk_arr : numpy.ndarray[dtype=float32]
        The 1D air temperature array (Kelvin)
    dwpk_arr : numpy.ndarray[dtype=float32]
        The 1D dewpoint temperature array (Kelvin)

    Returns
    -------
    numpy.ndarray[dtype=float32]
        The 1D array of LCL temperatures (Kelvin)
    """

@overload
def vapor_pressure(pressure: float, temperature: float) -> float:
    """
    Compute the vapor pressure with respect to liquid water.

    The vapor pressure (or saturation vapor pressure) is computed with 
    respect to liquid water when the dewpoint temperature (or air temperature)
    is passed (in Kelvin), along with the air pressure (Pa). The air pressure 
    is used as a minimum floor value for extremely cold temperatures at low
    pressures, and is consistent with how vapor pressure is treated in CM1. 

    This function uses the formulation by Bolton (1980), and is
    accurate to within 0.3% for the temperature range of -35C <= T <= 35C.

    Parameters
    ----------
    pressure : float 
        The air pressure (Pa)
    temperature : float 
        The air temperature (K) or dewpoint temperature (K)

    Returns
    -------
    float
        The vapor pressure (Pa) given dewpoint temperature, or the 
        saturation vapor pressure given the air temperature.
    """

@overload
def vapor_pressure(pres_arr: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], tmpk_arr: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')]:
    """
    Compute the vapor pressure with respect to liquid water.

    The vapor pressure (or saturation vapor pressure) is computed with 
    respect to liquid water when the dewpoint temperature (or air temperature)
    is passed (in Kelvin), along with the air pressure (Pa). The air pressure 
    is used as a minimum floor value for extremely cold temperatures at low
    pressures, and is consistent with how vapor pressure is treated in CM1. 

    This function uses the formulation by Bolton (1980), and is
    accurate to within 0.3% for the temperature range of -35C <= T <= 35C.

    Parameters
    ----------
    pres_arr : numpy.ndarray[dtype=float32] 
        The 1D array of air pressure (Pa)
    tmpk_arr : numpy.ndarray[dtype=float32]
        The 1D array of air temperature (K) or dewpoint temperature (K)

    Returns
    -------
    numpy.ndarray[dtype=float32]
        The 1D array of vapor pressure (Pa) given dewpoint temperature, or the 
        saturation vapor pressure given the air temperature.
    """

@overload
def vapor_pressure_ice(pressure: float, temperature: float) -> float:
    """
    Compute the vapor pressure with respect to ice.

    The vapor pressure (or saturation vapor pressure) is computed with 
    respect to ice when the dewpoint temperature (or air temperature)
    is passed (in Kelvin), along with the air pressure (Pa). The air pressure 
    is used as a minimum floor value for extremely cold temperatures at low
    pressures, and is consistent with how vapor pressure is treated in CM1. 

    This function uses the formulation by Bolton (1980), and is
    accurate to within 0.3% for the temperature range of -35C <= T <= 35C.

    Parameters
    ----------
    pressure : float 
        The air pressure (Pa)
    temperature : float 
        The air temperature (K) or dewpoint temperature (K)

    Returns
    -------
    float
        The vapor pressure (Pa) given dewpoint temperature, or the 
        saturation vapor pressure given the air temperature.
    """

@overload
def vapor_pressure_ice(pres_arr: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], tmpk_arr: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')]:
    """
    Compute the vapor pressure with respect to ice.

    The vapor pressure (or saturation vapor pressure) is computed with 
    respect to ice when the dewpoint temperature (or air temperature)
    is passed (in Kelvin), along with the air pressure (Pa). The air pressure 
    is used as a minimum floor value for extremely cold temperatures at low
    pressures, and is consistent with how vapor pressure is treated in CM1. 

    This function uses the formulation by Bolton (1980), and is
    accurate to within 0.3% for the temperature range of -35C <= T <= 35C.

    Parameters
    ----------
    pres_arr : numpy.ndarray[dtype=float32] 
        The 1D array of air pressure (Pa)
    tmpk_arr : numpy.ndarray[dtype=float32]
        The 1D array of air temperature (K) or dewpoint temperature (K)

    Returns
    -------
    numpy.ndarray[dtype=float32]
        The 1D array of vapor pressure (Pa) given dewpoint temperature, or the 
        saturation vapor pressure given the air temperature.
    """

@overload
def temperature_at_mixratio(wv_mixratio: float, pressure: float) -> float:
    """
    Computes the temperature (K) of air at the given water vapor mixing ratio
    (kg/kg) and air pressure (Pa). Can be used to compute the dewpoint temperature 
    from mixing ratio. 

    The routine is implemented as in Bolton (1980) eq 11, and is considered to be 
    accurate to 0.03 K for -35C <= T <= 35C. 

    Parameters
    ----------
    wv_mixratio : float 
        The water vapor mixing ratio (kg/kg)
    pressure : float 
        The air pressure (Pa)

    Returns
    -------
    float
        The temperature (K) of an air parcel at a given mixing ratio and pressure.
    """

@overload
def temperature_at_mixratio(mixr_arr: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], pres_arr: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')]:
    """
    Computes the temperature (K) of air at the given water vapor mixing ratio
    (kg/kg) and air pressure (Pa). Can be used to compute the dewpoint temperature 
    from mixing ratio. 

    The routine is implemented as in Bolton (1980) eq 11, and is considered to be 
    accurate to 0.03 K for -35C <= T <= 35C. 

    Parameters
    ----------
    mixr_arr : numpy.ndarray[dtype=float32] 
        The 1D array of water vapor mixing ratio (kg/kg)
    pres_arr : numpy.ndarray[dtype=float32]
        The 1D array of air pressure (Pa)

    Returns
    -------
    numpy.ndarray[dtype=float32]
        The 1D array of temperature (K) of an air parcel at a given mixing ratio and pressure.
    """

@overload
def theta_level(potential_temperature: float, temperature: float) -> float:
    """
    Computes the pressure level (Pa) of a parcel given the potential temperature (K) and air 
    temperature (K).

    Parameters
    ----------
    potential_temperature : float 
        The potential temperature, or theta (K)
    temperature : float 
        The air temperature (K)

    Returns
    -------
    float
        The pressure level (Pa) corresponding to the potential temperature and air temperature.
    """

@overload
def theta_level(theta_arr: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], tmpk_arr: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')]:
    """
    Computes the pressure level (Pa) of a parcel given the potential temperature (K) and air 
    temperature (K).

    Parameters
    ----------
    theta_arr : numpy.ndarray[dtype=float32] 
        The 1D array of potential temperature, or theta (K)
    tmpk_arr : numpy.ndarray[dtype=float32]
        The 1D array of air temperature (K)

    Returns
    -------
    numpy.ndarray[dtype=float32]
        The 1D pressure level (Pa) corresponding to the potential temperature and air temperature.
    """

@overload
def theta(pressure: float, temperature: float, ref_pressure: float = 100000.0) -> float:
    """
    Computes the potential temperature (K), or theta, given the air pressure (Pa), air temperature (K),
    and a reference pressure (default value is 100000 Pa).

    Parameters
    ----------
    pressure : float 
        The air pressure (Pa)
    temperature : float 
        The air temperature (K)
    ref_pressure : float, optional 
        The reference pressure (default 100000.0 Pa)

    Returns
    -------
    float
        The potential temperature (K), or theta
    """

@overload
def theta(pres_arr: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], tmpk_arr: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], ref_pressure: float = 100000.0) -> Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')]:
    """
    Computes the potential temperature (K), or theta, given the air pressure (Pa), air temperature (K),
    and a reference pressure (default value is 100000 Pa).

    Parameters
    ----------
    pressure : numpy.ndarray[dtype=float32]
        The 1D array of air pressure (Pa)
    temperature : numpy.ndarray[dtype=float32]
        The 1D array of air temperature (K)
    ref_pressure : float, optional
        The reference pressure (default 100000.0 Pa)

    Returns
    -------
    numpy.ndarray[dtype=float32]
        The 1D array of potential temperature (K), or theta
    """

@overload
def mixratio(q: float) -> float:
    """
    Compute the water vapor mixing ratio (kg/kg) from specific humidity (kg/kg).

    Parameters
    ----------
    q : float 
        The specific humidity (kg/kg)

    Returns
    -------
    float
        The water vapor mixing ratio (kg/kg)
    """

@overload
def mixratio(pressure: float, temperature: float) -> float:
    """
    Compute the water vapor mixing ratio (kg/kg) from the air pressure (Pa) and temperature (K).
    If given the air temperature, this is the saturation mixing ratio. If given the dewpoint 
    temperature, tis is the mixing ratio. 

    Parameters
    ----------
    pressure : float
        The air pressure (Pa)
    temperature : float 
        The air temperature (K)

    Returns
    -------
    float
        The water vapor mixing ratio (kg/kg)
    """

@overload
def mixratio(spfh_arr: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')]:
    """
    Compute the water vapor mixing ratio (kg/kg) from the specific humidity (kg/kg).

    Parameters
    ----------
    spfh_arr : numpy.ndarray[dtype=float32] 
        The 1D array of specific humidity (kg/kg)

    Returns
    -------
    numpy.ndarray[dtype=float32]
        The 1D array of water vapor mixing ratio (kg/kg)
    """

@overload
def mixratio(pres_arr: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], tmpk_arr: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')]:
    """
    Compute the water vapor mixing ratio (kg/kg) from the air pressure (Pa) and temperature (K).
    If given the air temperature, this is the saturation mixing ratio. If given the dewpoint 
    temperature, this is the mixing ratio. 

    Parameters
    ----------
    pres_arr : numpy.ndarray[dtype=float32] 
        The 1D array of air pressure (Pa)
    tmpk_arr : numpy.ndarray[dtype=float32]
        The 1D array of air temperature (K)

    Returns
    -------
    numpy.ndarray[dtype=float32]
        The 1D array of water vapor mixing ratio (kg/kg)
    """

@overload
def mixratio_ice(pressure: float, temperature: float) -> float:
    """
    Compute the ice water mixing ratio (kg/kg) from the air pressure (Pa) and temperature (K).
    If given the air temperatuer, this is the saturation mixing ratio. If given the dewpoint 
    temperature, this is the mixing ratio. 

    Parameters
    ----------
    pressure : float 
        The air pressure (Pa)
    temperature : float 
        The air temperature (K)

    Returns
    -------
    float
        The ice water mixing ratio (kg/kg)
    """

@overload
def mixratio_ice(pres_arr: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], tmpk_arr: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')]:
    """
    Compute the ice water mixing ratio (kg/kg) from the air pressure (Pa) and temperature (K).
    If given the air temperatuer, this is the saturation mixing ratio. If given the dewpoint 
    temperature, this is the mixing ratio. 

    Parameters
    ----------
    pres_arr : numpy.ndarray[dtype=float32]
        The 1D array of air pressure (Pa)
    tmpl_arr : numpy.ndarray[dtype=float32] 
        The 1D array of air temperature (K)

    Returns
    -------
    numpy.ndarray[dtype=float32]
        The ice water mixing ratio (kg/kg)
    """

@overload
def specific_humidity(rv: float) -> float:
    """
    Compute the specific humidity (kg/kg) from a mixing ratio (kg/kg).

    Parameters
    ----------
    rv : float 
        The water vapor mixing ratio (kg/kg)

    Returns
    -------
    float
        The specific humidity (kg/kg)
    """

@overload
def specific_humidity(mixr_arr: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')]:
    """
    Compute the specific humidity (kg/kg) from a mixing ratio (kg/kg).

    Parameters
    ----------
    mixr_arr : numpy.ndarray[dtype=float32]
        The 1D array of water vapor mixing ratios (kg/kg)

    Returns
    -------
    numpy.ndarray[dtype=float32]
        1D NumPy array of specific humidity (unitless)
    """

@overload
def relative_humidity(pressure: float, temperature: float, dewpoint: float) -> float:
    """
    Compute the relative humidity of water vapor with respect to liquid water.
    NOTE: The pressure variable is only used as a sanity check when computing 
    vapor pressure at extremely low pressures and temperatures. If you do not 
    want or need this behavior, you can pass the THETA_REF_PRESSURE constant
    in place of an air pressure.

    Parameters
    ----------
    pressure : float 
        the ambient air pressure (Pa)
    temperature : float 
        the ambient air temperature (K)
    dewpoint : float 
        the ambient dewpoint temperature (K)

    Returns
    -------
    float
        Relative Humidity (fraction, unitless)
    """

@overload
def relative_humidity(pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], dewpoint: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')]:
    """
    Compute the relative humidity of water vapor with respect to liquid water.
    NOTE: The pressure variable is only used as a sanity check when computing 
    vapor pressure at extremely low pressures and temperatures. If you do not
    want or need this behavior, you canpass in an array of THETA_REF_PRESSURE
    in place of air pressure.

    Parameters
    ----------
    pressure : numpy.ndarray[dtype=float32] 
        1D NumPy array of ambient air pressure Pa)
    temperature : numpy.ndarray[dtype=float32]
        1D NumPy array of ambient air temperature (K)
    dewpoint : numpy.ndarray[dtype=float32]
        1D NumPy array of ambient dewpoint temperature (K)

    Returns
    -------
    numpy.ndarray[dtype=float32]
        1D NumPy array of relative humidity values (fraction, unitless)
    """

@overload
def virtual_temperature(temperature: float, rv: float, rl: float = 0.0, ri: float = 0.0) -> float:
    """
    Returns the virtual temperature in Kelvin given the dry-bulb 
    temperature (Kelvin), the water vapor mixing ratio (kg/kg), the 
    liquid water mixing ratio (kg/kg), and the ice water mixing ratios
    (kg/kg). The liquid and ice water mixing ratios have default values 
    of zero, if unspecified. 

    Parameters
    ----------
    temperature : float 
        The dry-bulb temperature (K)
    rv : float 
        The water vapor mixing ratio (kg/kg)
    rl : float 
        The liquid water mixing ratio (kg/kg) 
    ri : float 
        The ice water mixing ratio (kg/kg)

    Returns
    -------
    float
        The virtual temperature (K)
    """

@overload
def virtual_temperature(tmpk_arr: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], rv_arr: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], rl_arr: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)] | None = None, ri_arr: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)] | None = None) -> Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')]:
    """
    Returns the virtual temperature in Kelvin given the dry-bulb 
    temperature (Kelvin), the water vapor mixing ratio (kg/kg), the 
    liquid water mixing ratio (kg/kg), and the ice water mixing ratios
    (kg/kg). The liquid and ice water mixing ratios have default values 
    of zero, if unspecified. 

    Parameters
    ----------
    tmpk_arr : numpy.ndarray[dtype=float32] 
        The 1D array of dry-bulb temperature (K)
    rv_arr : numpy.ndarray[dtype=float32] 
        The 1D array of water vapor mixing ratio (kg/kg)
    rl_arr : numpy.ndarray[dtype=float32], optional 
        The 1D array of liquid water mixing ratio (kg/kg) 
    ri_arr : numpy.ndarray[dtype=float32], optional 
        The 1D array of ice water mixing ratio (kg/kg)

    Returns
    -------
    numpy.ndarray[dtype=float32]
        The 1D array of virtual temperature (K)
    """

@overload
def wetlift(pressure: float, temperature: float, lifted_pressure: float, converge: float = 0.0010000000474974513) -> float:
    """
    Compute the temperature of a parcel lifted moist adiabatically to a new level. 

    With a given parcel defined by a pressure (Pa) and temperature (K), lift it 
    moist adiabatically to a new pressure level (Pa) and return the temperature of 
    the parcel at that level. 

    This function relies on the Wobus Function (thermo.wobf), and it was shown by 
    Robert Davies-Jones (2007) that the WObus function has a slight dependence on 
    pressure, which results in errors of up to 1.2 K in the temperature of a lifted 
    parcel. 

    Parameters
    ----------
    pressure : float 
        The air pressure (Pa)
    temperature : float 
        The saturated air temperature (K)
    lifted_pressure : float 
        The new pressure level to lift to (Pa)
    converge : float 
        The iterative convergence criteria (K; default = 0.001)

    Returns
    -------
    float
        The new temperature (K) when lifted moist adiabatically to the new pressure level
    """

@overload
def wetlift(arg0: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], arg1: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], arg2: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], /) -> Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')]:
    """
    Compute the temperature of a parcel lifted moist adiabatically to a new level. 

    With a given parcel defined by a pressure (Pa) and temperature (K), lift it 
    moist adiabatically to a new pressure level (Pa) and return the temperature of 
    the parcel at that level. 

    This function relies on the Wobus Function (thermo.wobf), and it was shown by 
    Robert Davies-Jones (2007) that the WObus function has a slight dependence on 
    pressure, which results in errors of up to 1.2 K in the temperature of a lifted 
    parcel. 

    Parameters
    ----------
    pressure : numpy.ndarray[dtype=float32] 
        The 1D array of air pressures (Pa)
    temperature : numpy.ndarray[dtype=float32] 
        The 1D array of saturated air temperatures (K)
    lifted_pressure : numpy.ndarray[dtype=float32]
        The 1D array of new pressure levels to lift to (Pa)

    Returns
    -------
    numpy.ndarray[dtype=float32]
        The 1D array of new temperatures (K) when lifted moist adiabatically to the new pressure levels
    """

def drylift(pressure: float, temperature: float, dewpoint: float) -> tuple[float, float]:
    """
    Given a parcel's initial pressure (Pa), temperature (K), and 
    dewpoint temperature (K), lift the parcel dry adiabaitically to its
    Lifted Condensation Level and return the resulting LCL pressure (Pa)
    and temperature (K).

    The LCL temperature is computed using an approximation. See the 
    lcl_temperature function documentation for more information.

    Parameters
    ----------
    pressure : float 
        Parcel starting pressure (Pa)
    temperature : float 
        Parcel starting temperature (K)
    dewpoint : float 
        Parcel starting dewpoint temperature (K)

    Returns
    -------
    tuple[float, float]
        A tuple of (lcl_pres, lcl_temperature)
    """

@overload
def wetbulb(lifter: nwsspc.sharp.calc.parcel.lifter_wobus, pressure: float, temperature: float, dewpoint: float) -> float:
    """
    Compute the wet bulb temperature (K) given the ambient pressure (Pa), 
    temperature (K), and dewpoint temperature (K).

    First, it lifts a parcel with the given pressure, temperature, and
    dewpoint temperature to its Lifted Condensation Level (LCL). To compute the 
    temperature and pressure of the LCL, an approximation is used. See the 
    lcl_temperature function for further detail. 

    After the parcel has reached the LCL, the lifter passed to the function 
    lowers the parcel to its initial pressure level along a moist adiabat or 
    pseudoadiabat. 

    Parameters
    ----------
    lifter : nwsspc.sharp.calc.parcel.lifter_wobus 
        a parcel lifter (e.g. lifter_cm1 or lifter_wobus)
    pressure : float 
        The ambient pressure (Pa)
    temperature : float 
        The ambient temperature (K)
    dewpoint : float 
        The ambient dewpoint temperature (K)

    Returns
    -------
    float
        The wetbulb temperature (K)
    """

@overload
def wetbulb(lifter: nwsspc.sharp.calc.parcel.lifter_cm1, pressure: float, temperature: float, dewpoint: float) -> float:
    """
    Compute the wet bulb temperature (K) given the ambient pressure (Pa), 
    temperature (K), and dewpoint temperature (K).

    First, it lifts a parcel with the given pressure, temperature, and
    dewpoint temperature to its Lifted Condensation Level (LCL). To compute the 
    temperature and pressure of the LCL, an approximation is used. See the 
    lcl_temperature function for further detail. 

    After the parcel has reached the LCL, the lifter passed to the function 
    lowers the parcel to its initial pressure level along a moist adiabat or 
    pseudoadiabat. 

    Parameters
    ----------
    lifter : nwsspc.sharp.calc.parcel.lifter_cm1 
        a parcel lifter (e.g. lifter_cm1 or lifter_wobus)
    pressure : float 
        The ambient pressure (Pa)
    temperature : float 
        The ambient temperature (K)
    dewpoint : float 
        The ambient dewpoint temperature (K)

    Returns
    -------
    float
        The wetbulb temperature (K)
    """

@overload
def wetbulb(lifter: nwsspc.sharp.calc.parcel.lifter_wobus, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], dewpoint: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')]:
    """
    Compute the wet bulb temperature (K) given the ambient pressure (Pa), 
    temperature (K), and dewpoint temperature (K).

    First, it lifts a parcel with the given pressure, temperature, and
    dewpoint temperature to its Lifted Condensation Level (LCL). To compute the 
    temperature and pressure of the LCL, an approximation is used. See the 
    lcl_temperature function for further detail. 

    After the parcel has reached the LCL, the lifter passed to the function 
    lowers the parcel to its initial pressure level along a moist adiabat or 
    pseudoadiabat. 

    Parameters
    ----------
    lifter : nwsspc.sharp.calc.parcel.lifter_wobus 
        a parcel lifter (e.g. lifter_cm1 or lifter_wobus)
    pressure : numpy.ndarray[dtype=float32] 
        1D NumPy array of ambient pressures (Pa)
    temperature : numpy.ndarray[dtype=float32] 
        1D NumPy array of ambient temperatureds (K)
    dewpoint : numpy.ndarray[dtype=float32] 
        1D NumPy array of ambient dewpoint temperatures (K)

    Returns
    -------
    numpy.ndarray[dtype=float32]
        1D NumPy array of wetbulb temperatures (K)
    """

@overload
def wetbulb(lifter: nwsspc.sharp.calc.parcel.lifter_cm1, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], dewpoint: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')]:
    """
    Compute the wet bulb temperature (K) given the ambient pressure (Pa), 
    temperature (K), and dewpoint temperature (K).

    First, it lifts a parcel with the given pressure, temperature, and
    dewpoint temperature to its Lifted Condensation Level (LCL). To compute the 
    temperature and pressure of the LCL, an approximation is used. See the 
    lcl_temperature function for further detail. 

    After the parcel has reached the LCL, the lifter passed to the function 
    lowers the parcel to its initial pressure level along a moist adiabat or 
    pseudoadiabat. 

    Parameters
    ----------
    lifter : nwsspc.sharp.calc.parcel.lifter_cm1 
        a parcel lifter (e.g. lifter_cm1 or lifter_wobus)
    pressure : numpy.ndarray[dtype=float32] 
        1D NumPy array of ambient pressures (Pa)
    temperature : numpy.ndarray[dtype=float32] 
        1D NumPy array of ambient temperatureds (K)
    dewpoint : numpy.ndarray[dtype=float32] 
        1D NumPy array of ambient dewpoint temperatures (K)

    Returns
    -------
    numpy.ndarray[dtype=float32]
        1D NumPy array of wetbulb temperatures (K)
    """

@overload
def theta_wetbulb(lifter: nwsspc.sharp.calc.parcel.lifter_wobus, pressure: float, temperature: float, dewpoint: float) -> float:
    """
    Compute the wet-bulb potential temperature (K) given the pressure (Pa), 
    temperature (K), and dewpoint temperature (K).

    First, it lifts a parcel with the given pressure, temperature, 
    and dewpoint temperature to its Lifted Condensation Level (LCL). 
    To compute the temperature and pressure of the LCL, an approximation 
    is used. See the lcl_temperature fuction for further detail. 

    After the parcel has reached the LCL, the lifted passed to the function 
    lowers the parcel to the standard parcel reference pressure level 
    (1000 hPa) along a moist adiabat.

    Parameters
    ----------
    lifter : nwsspc.sharp.calc.parcel.lifter_wobus 
        a parcel lifter (e.g. lifter_cm1 or lifter_wobus)
    pressure : float 
        the ambient air pressure (Pa)
    temperature : float 
        the ambient air temperature (K)
    dewpoint : float 
        the ambient dewpoint temperature

    Returns
    -------
    float
        The wet-bulb potential temperature (K)
    """

@overload
def theta_wetbulb(lifter: nwsspc.sharp.calc.parcel.lifter_cm1, pressure: float, temperature: float, dewpoint: float) -> float:
    """
    Compute the wet-bulb potential temperature (K) given the pressure (Pa), 
    temperature (K), and dewpoint temperature (K).

    First, it lifts a parcel with the given pressure, temperature, 
    and dewpoint temperature to its Lifted Condensation Level (LCL). 
    To compute the temperature and pressure of the LCL, an approximation 
    is used. See the lcl_temperature fuction for further detail. 

    After the parcel has reached the LCL, the lifted passed to the function 
    lowers the parcel to the standard parcel reference pressure level 
    (1000 hPa) along a moist adiabat.

    Parameters
    ----------
    lifter : nwsspc.sharp.calc.parcel.lifter_cm1 
        a parcel lifter (e.g. lifter_cm1 or lifter_wobus)
    pressure : float 
        the ambient air pressure (Pa)
    temperature : float 
        the ambient air temperature (K)
    dewpoint : float 
        the ambient dewpoint temperature

    Returns
    -------
    float
        The wet-bulb potential temperature (K)
    """

@overload
def theta_wetbulb(lifter: nwsspc.sharp.calc.parcel.lifter_wobus, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], dewpoint: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')]:
    """
    Compute the wet-bulb potential temperature (K) given the pressure (Pa), 
    temperature (K), and dewpoint temperature (K).

    First, it lifts a parcel with the given pressure, temperature, 
    and dewpoint temperature to its Lifted Condensation Level (LCL). 
    To compute the temperature and pressure of the LCL, an approximation 
    is used. See the lcl_temperature fuction for further detail. 

    After the parcel has reached the LCL, the lifted passed to the function 
    lowers the parcel to the standard parcel reference pressure level 
    (1000 hPa) along a moist adiabat.

    Parameters
    ----------
    lifter : nwsspc.sharp.calc.parcel.lifter_wobus 
        a parcel lifter (e.g. lifter_cm1 or lifter_wobus)
    pressure : numpy.ndarray[dtype=float32] 
        1D NumPy array of ambient air pressure (Pa)
    temperature : numpy.ndarray[dtype=float32] 
        1D NumPy array of ambient air temperature (K)
    dewpoint : numpy.ndarray[dtype=float32] 
        1D NumPy array of ambient dewpoint temperature

    Returns
    -------
    numpy.ndarray[dtype=float32]
        1D NumPy array of wet-bulb potential temperatures (K)
    """

@overload
def theta_wetbulb(lifter: nwsspc.sharp.calc.parcel.lifter_cm1, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], dewpoint: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')]:
    """
    Compute the wet-bulb potential temperature (K) given the pressure (Pa), 
    temperature (K), and dewpoint temperature (K).

    First, it lifts a parcel with the given pressure, temperature, 
    and dewpoint temperature to its Lifted Condensation Level (LCL). 
    To compute the temperature and pressure of the LCL, an approximation 
    is used. See the lcl_temperature fuction for further detail. 

    After the parcel has reached the LCL, the lifted passed to the function 
    lowers the parcel to the standard parcel reference pressure level 
    (1000 hPa) along a moist adiabat.

    Parameters
    ----------
    lifter : nwsspc.sharp.calc.parcel.lifter_cm1 
        a parcel lifter (e.g. lifter_cm1 or lifter_wobus)
    pressure : numpy.ndarray[dtype=float32] 
        1D NumPy array of ambient air pressure (Pa)
    temperature : numpy.ndarray[dtype=float32] 
        1D NumPy array of ambient air temperature (K)
    dewpoint : numpy.ndarray[dtype=float32] 
        1D NumPy array of ambient dewpoint temperature

    Returns
    -------
    numpy.ndarray[dtype=float32]
        1D NumPy array of wet-bulb potential temperatures (K)
    """

@overload
def thetae(pressure: float, temperature: float, dewpoint: float) -> float:
    """
    Compute the equivalent potential temperature (K) given the 
    air pressure (Pa), air temperature (K), and dewpoint temperature (K).
    The equivalent potential temperature is computed as in Bolton 1980. 
    It was found in Davies-Jones 2009 to be the most accurate non-iterative 
    formulation of theta-e.

    References
    ----------
    Bolton 1980: https://doi.org/10.1175/1520-0493(1980)108%3C1046:TCOEPT%3E2.0.CO;2

    Davies-Jones 2009: https://doi.org/10.1175/2009MWR2774.1

    Parameters
    ----------
    pressure : float 
        the ambient air pressure (Pa)
    temperature : float 
        the ambient air temperature (K)
    dewpoint : float 
        the dewpoint temperature (K)

    Returns
    -------
    float
        The equivalent potential temperature (K)
    """

@overload
def thetae(pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], dewpoint: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')]:
    """
    Compute the equivalent potential temperature (K) given the 
    air pressure (Pa), air temperature (K), and dewpoint temperature (K).
    The equivalent potential temperature is computed as in Bolton 1980. 
    It was found in Davies-Jones 2009 to be the most accurate non-iterative 
    formulation of theta-e.

    References 
    ----------
    Bolton 1980: https://doi.org/10.1175/1520-0493(1980)108%3C1046:TCOEPT%3E2.0.CO;2

    Davies-Jones 2009: https://doi.org/10.1175/2009MWR2774.1

    Parameters
    ----------
    pressure : numpy.ndarray[dtype=float32] 
        1D NumPy array of ambient air pressure (Pa)
    temperature : numpy.ndarray[dtype=float32] 
        1D NumPy array of ambient air temperature (K)
    dewpoint : numpy.ndarray[dtype=float32] 
        1D NumPy array of dewpoint temperature (K)

    Returns
    -------
    numpy.ndarray[dtype=float32]
        1D NumPy array of equivalent potential temperature (K)
    """

@overload
def lapse_rate(layer_agl: nwsspc.sharp.calc.layer.HeightLayer, height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> float:
    """
    Computes the lapse rate over a given HeightLayer (meters AGL).
    This routine handles converting the height AGL to MSL by adding
    the surface height value to the HeightLayer.

    Parameters
    ----------
    layer_agl : nwsspc.sharp.calc.layer.HeightLayer 
        a HeightLayer (meters AGL) to compute the lapse rate over
    height : numpy.ndarray[dtype=float32] 
        1D NumPy array of height values (meters)
    temperature : numpy.ndarray[dtype=float32] 
        1D NumPy array of temperature values (K)

    Returns
    -------
    float
        The temperature lapse rate (K)
    """

@overload
def lapse_rate(layer: nwsspc.sharp.calc.layer.PressureLayer, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> float:
    """
    Computes the lapse rate over a given PressureLayer (Pa).

    Parameters
    ----------
    layer : nwsspc.sharp.calc.layer.PressureLayer 
        a PressureLayer (Pa) to compute the lapse rate over
    pressure : numpy.ndarray[dtype=float32] 
        1D NumPy array of pressure values (Pa)
    height : numpy.ndarray[dtype=float32] 
        1D NumPy array of height values (meters)
    temperature : numpy.ndarray[dtype=float32] 
        1D NumPy array of temperature values (K)

    Returns
    -------
    float
        The temperature lapse rate (K)
    """

@overload
def lapse_rate_max(layer: nwsspc.sharp.calc.layer.HeightLayer, depth: float, height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> tuple[float, nwsspc.sharp.calc.layer.HeightLayer]:
    """
    Given a layer of the atmosphere (e.g. 2 - 6 km), find the maximum
    lapse rate over the provided depth (e.g. 2 km) within that given layer. 
    Returns the maximum lapse rate, as well as the layer it was found in. 

    Parameters
    ----------
    layer : nwsspc.sharp.calc.layer.HeightLayer 
        a HeightLayer to search for the maximum lapse lapse_rate (meters AGL) 
    depth : float 
        the depth used to compute a lapse rate within a layer (meters)
    height : numpy.ndarray[dtype=float32] 
        1D NumPy array of height values (meters)
    temperature : numpy.ndarray[dtype=float32] 
        1D NumPy array of temperature values (K)

    Returns
    -------
    tuple[float, nwsspc.sharp.calc.layer.HeightLayer]
        A tuple containing the maximum lapse rate and the layer it was found in
    """

@overload
def lapse_rate_max(layer: nwsspc.sharp.calc.layer.PressureLayer, depth: float, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> tuple[float, nwsspc.sharp.calc.layer.PressureLayer]:
    """
    Given a layer of the atmosphere (e.g. 800 hPa - 500 hPa), find the maximum
    lapse rate over the provided depth (e.g. 100 hPa) within that given layer. 
    Returns the maximum lapse rate, as well as the layer it was found in. 

    Parameters
    ----------
    layer : nwsspc.sharp.calc.layer.PressureLayer 
        a PressureLayer to search for the maximum lapse lapse_rate (Pa) 
    depth : float 
        the depth used to compute a lapse rate within a layer (Pa)
    pressure : numpy.ndarray[dtype=float32] 
        1D NumPy array of pressure values (Pa)
    height : numpy.ndarray[dtype=float32] 
        1D NumPy array of height values (meters)
    temperature : numpy.ndarray[dtype=float32] 
        1D NumPy array of temperature values (K)

    Returns
    -------
    tuple[float, nwsspc.sharp.calc.layer.PressureLayer]
        A tuple containing the maximum lapse rate and the layer it was found in
    """

@overload
def buoyancy(parcel_temperature: float, environment_temperature: float) -> float:
    """
    Compute buoyancy given parcel & environment temperatures.

    Parameters
    ----------
    parcel_temperature : float 
        parcel temperature (K)
    environment_temperature : float 
        environment temperature (K)

    Returns
    -------
    float
        Buoyancy (m/s^2)
    """

@overload
def buoyancy(parcel_temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], environment_temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C')]:
    """
    Compute buoyancy given arrays of parcel & environment temperatures.

    Parameters
    ----------
    parcel_temperature : numpy.ndarray[dtype=float32] 
        1D NumPy array of parcel temperatures (K)
    environment_temperature : numpy.ndarray[dtype=float32] 
        1D NumPy array of environment temperatures (K)

    Returns
    -------
    numpy.ndarray[dtype=float32]
        1D NumPy array of buoyancy values (m/s^2)
    """

def pbl_top(pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], thetav: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], offset: float = 0.5) -> int:
    """
    Compute the pressure of the top of the Planetary Boundary Layer (PBL). 
    Uses the method described by Stull (1988), by which the virtual potential 
    temperature is used. The offset determines the termination condition above 
    the surface.

    Parameters 
    ----------
    pressure : numpy.ndarray[dtype=float32] 
        1D NumPy array of pressure values (Pa)
    thetav : numpy.ndarray[dtype=float32] 
        1D NumPy array of virtual potential temperature values (K)
    offset : float 
        The virtual potential temperature offset (K)

    Returns 
    -------
    float
        The pressure of the PBL top (Pa)
    """

def temperature_layer(pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], temperature: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], tmpk_lo: float, tmpk_hi: float, pres_min: float = 10000.0) -> nwsspc.sharp.calc.layer.PressureLayer:
    """
    Performs a top-down search for a PressureLayer that bounds 
    the given temperature range. Uses a geometric approach to 
    find the intersection of the temperature line with the given range.

    Parameters 
    ----------
    pressure : numpy.ndarray[dtype=float32] 
        1D NumPy array of pressure values (Pa)
    temperature : numpy.ndarray[dtype=float32] 
        1D NumPy array of temperature values (K)
    tmpk_lo : float 
        The low end of the temperature range (K)
    tmpk_hi : float 
        The high end of the temperature range (K)
    pres_min : float 
        Restricts the search pressure levels (Pa)

    Returns 
    -------
    nwsspc.sharp.calc.layer.PressureLayer 
        The layer bounding the temperature range from a top-down search
    """




import os
import pytest
import numpy as np
import pandas as pd

from nwsspc.sharp.calc import layer
from nwsspc.sharp.calc import parcel
from nwsspc.sharp.calc import thermo
from nwsspc.sharp.calc import constants


def load_parquet(filename):
    snd_df = pd.read_parquet(filename)
    snd_df = snd_df[snd_df["vwin"].notna()]
    snd_df = snd_df[snd_df["tmpc"].notna()]
    snd_df = snd_df[snd_df["relh"].notna()]
    snd_df = snd_df[snd_df["pres"] >= 50.0]

    pres = snd_df["pres"].to_numpy().astype('float32')*np.float32(100.0)
    hght = snd_df["hght"].to_numpy().astype('float32')
    tmpk = snd_df["tmpc"].to_numpy().astype('float32')+np.float32(273.15)
    dwpk = snd_df["dwpc"].to_numpy().astype('float32')+np.float32(273.15)
    wdir = snd_df["wdir"].to_numpy().astype('float32')
    wspd = snd_df["wspd"].to_numpy().astype('float32')
    uwin = snd_df["uwin"].to_numpy().astype('float32')
    vwin = snd_df["vwin"].to_numpy().astype('float32')

    # turn into height above ground level
    hght -= hght[0]

    mixr = thermo.mixratio(pres, dwpk)
    vtmp = thermo.virtual_temperature(tmpk, mixr)
    thetae = thermo.thetae(
        pres,
        tmpk,
        dwpk
    )

    return {
        "pres": pres, "hght": hght,
        "tmpk": tmpk, "mixr": mixr,
        "vtmp": vtmp, "dwpk": dwpk,
        "thetae": thetae,
        "wdir": wdir, "wspd": wspd,
        "uwin": uwin, "vwin": vwin
    }


data_dir = os.path.join(
    os.path.dirname(__file__),
    "..",
    "..",
    "data",
    "test_snds"
)
filename = os.path.join(
    data_dir,
    "ddc.parquet"
)
snd_data = load_parquet(filename)


def test_lifter_wobus():
    lifter = parcel.lifter_wobus()

    pres = 100000.0
    tmpk = 320.0
    new_pres = 50000.0

    assert (lifter(constants.MISSING, tmpk, new_pres) == constants.MISSING)
    assert (lifter(pres, constants.MISSING, new_pres) == constants.MISSING)
    assert (lifter(pres, tmpk, constants.MISSING) == constants.MISSING)

    assert (np.isnan(lifter(np.nan, tmpk, new_pres)))
    assert (np.isnan(lifter(pres, np.nan, new_pres)))
    assert (np.isnan(lifter(pres, tmpk, np.nan)))

    assert (lifter(pres, tmpk, new_pres) == pytest.approx(301.872528))


def test_lifter_cm1():
    lifter = parcel.lifter_cm1()
    lifter.ma_type = thermo.adiabat.pseudo_liq
    lifter.converge = 100.0

    pres = 100000.0
    tmpk = 320.0
    new_pres = 50000.0

    assert (lifter(constants.MISSING, tmpk, new_pres) == constants.MISSING)
    assert (lifter(pres, constants.MISSING, new_pres) == constants.MISSING)
    assert (lifter(pres, tmpk, constants.MISSING) == constants.MISSING)

    assert (np.isnan(lifter(np.nan, tmpk, new_pres)))
    assert (np.isnan(lifter(pres, np.nan, new_pres)))


def test_surface_parcel():
    pres = snd_data["pres"][0]
    tmpk = snd_data["tmpk"][0]
    dwpk = snd_data["dwpk"][0]

    pcl = parcel.Parcel.surface_parcel(pres, tmpk, dwpk)

    # Wobus Lifter
    lifter = parcel.lifter_wobus()
    vtmpk = pcl.lift_parcel(lifter, snd_data["pres"])
    buoy = thermo.buoyancy(vtmpk, snd_data["vtmp"])
    cape, cinh = pcl.cape_cinh(snd_data["pres"], snd_data["hght"], buoy)
    mpl = pcl.maximum_parcel_level(snd_data["pres"], snd_data["hght"], buoy)
    li = pcl.lifted_index(50000.0, snd_data["pres"], snd_data["vtmp"], vtmpk)

    assert (li == pytest.approx(-9.68, abs=1e-2))
    assert (cape == pytest.approx(3353.6, abs=1e-1))
    assert (cinh == pytest.approx(-34.5697, abs=1e-1))
    assert (pcl.lfc_pressure == pytest.approx(71729, abs=1e-0))
    assert (pcl.eql_pressure == pytest.approx(17933, abs=1e-0))
    assert (pcl.mpl_pressure == pytest.approx(9065, abs=1e-0))

    lifter = parcel.lifter_cm1()
    lifter.ma_type = thermo.adiabat.pseudo_liq
    vtmpk = pcl.lift_parcel(lifter, snd_data["pres"])
    buoy = thermo.buoyancy(vtmpk, snd_data["vtmp"])
    cape, cinh = pcl.cape_cinh(snd_data["pres"], snd_data["hght"], buoy)
    mpl = pcl.maximum_parcel_level(snd_data["pres"], snd_data["hght"], buoy)
    li = pcl.lifted_index(50000.0, snd_data["pres"], snd_data["vtmp"], vtmpk)

    assert (li == pytest.approx(-9.12, abs=1e-2))
    assert (cape == pytest.approx(3107.6, abs=5e-1))
    assert (cinh == pytest.approx(-36.4, abs=5e-1))
    assert (pcl.lfc_pressure == pytest.approx(71482, abs=1e-0))
    assert (pcl.eql_pressure == pytest.approx(18969, abs=1e-0))
    assert (pcl.mpl_pressure == pytest.approx(9529, abs=1e-0))


def test_mixed_layer_parcel():
    mix_lyr = layer.PressureLayer(
        snd_data["pres"][0], snd_data["pres"][0] - 10000.0)
    theta = thermo.theta(snd_data["pres"], snd_data["tmpk"])
    pcl = parcel.Parcel.mixed_layer_parcel(
        mix_lyr, snd_data["pres"], theta, snd_data["mixr"])

    assert (pcl.pres == pytest.approx(92043.0))
    assert (pcl.tmpk == pytest.approx(297.2543))
    assert (pcl.dwpk == pytest.approx(289.40469))

    lifter = parcel.lifter_wobus()
    vtmpk = pcl.lift_parcel(lifter, snd_data["pres"])
    buoy = thermo.buoyancy(vtmpk, snd_data["vtmp"])
    cape, cinh = pcl.cape_cinh(snd_data["pres"], snd_data["hght"], buoy)
    mpl = pcl.maximum_parcel_level(snd_data["pres"], snd_data["hght"], buoy)
    li = pcl.lifted_index(50000.0, snd_data["pres"], snd_data["vtmp"], vtmpk)

    assert (li == pytest.approx(-6.98, abs=1e-2))
    assert (cape == pytest.approx(2148.9, abs=1e-1))
    assert (cinh == pytest.approx(-128.49, abs=1e-1))
    assert (pcl.lfc_pressure == pytest.approx(67505, abs=1e-0))
    assert (pcl.eql_pressure == pytest.approx(20371, abs=1e-0))
    assert (pcl.mpl_pressure == pytest.approx(11080, abs=1e-0))

    lifter = parcel.lifter_cm1()
    vtmpk = pcl.lift_parcel(lifter, snd_data["pres"])
    buoy = thermo.buoyancy(vtmpk, snd_data["vtmp"])
    cape, cinh = pcl.cape_cinh(snd_data["pres"], snd_data["hght"], buoy)
    mpl = pcl.maximum_parcel_level(snd_data["pres"], snd_data["hght"], buoy)
    li = pcl.lifted_index(50000.0, snd_data["pres"], snd_data["vtmp"], vtmpk)

    assert (li == pytest.approx(-6.428, abs=1e-2))
    assert (cape == pytest.approx(1929.36, abs=5e-1))
    assert (cinh == pytest.approx(-133.71, abs=5e-1))
    assert (pcl.lfc_pressure == pytest.approx(67100, abs=1e-0))
    assert (pcl.eql_pressure == pytest.approx(20968, abs=1e-0))
    assert (pcl.mpl_pressure == pytest.approx(11649, abs=1e-0))


def test_most_unstable_parcel():
    search_layer = layer.PressureLayer(
        snd_data["pres"][0],
        snd_data["pres"][0] - 40000.0
    )
    lifter = parcel.lifter_wobus()

    pcl = parcel.Parcel.most_unstable_parcel(
        search_layer,
        lifter,
        snd_data["pres"],
        snd_data["hght"],
        snd_data["tmpk"],
        snd_data["vtmp"],
        snd_data["dwpk"]
    )
    vtmpk = pcl.lift_parcel(lifter, snd_data["pres"])
    buoy = thermo.buoyancy(vtmpk, snd_data["vtmp"])
    pcl.find_lfc_el(snd_data["pres"], snd_data["hght"], buoy)
    mpl = pcl.maximum_parcel_level(snd_data["pres"], snd_data["hght"], buoy)
    li = pcl.lifted_index(50000.0, snd_data["pres"], snd_data["vtmp"], vtmpk)

    assert (li == pytest.approx(-9.68, abs=1e-2))
    assert (pcl.pres == pytest.approx(92043.0))
    assert (pcl.tmpk == pytest.approx(298.15))
    assert (pcl.dwpk == pytest.approx(291.532))
    assert (pcl.cape == pytest.approx(3353.6, abs=1e-1))
    assert (pcl.cinh == pytest.approx(-34.5697, abs=1e-1))
    assert (pcl.lfc_pressure == pytest.approx(71729, abs=1e-0))
    assert (pcl.eql_pressure == pytest.approx(17933, abs=1e-0))
    assert (pcl.mpl_pressure == pytest.approx(9065, abs=1e-0))

    lifter = parcel.lifter_cm1()
    lifter.ma_type = thermo.adiabat.pseudo_liq

    pcl = parcel.Parcel.most_unstable_parcel(
        search_layer,
        lifter,
        snd_data["pres"],
        snd_data["hght"],
        snd_data["tmpk"],
        snd_data["vtmp"],
        snd_data["dwpk"]
    )
    vtmpk = pcl.lift_parcel(lifter, snd_data["pres"])
    buoy = thermo.buoyancy(vtmpk, snd_data["vtmp"])
    pcl.find_lfc_el(snd_data["pres"], snd_data["hght"], buoy)
    mpl = pcl.maximum_parcel_level(snd_data["pres"], snd_data["hght"], buoy)
    li = pcl.lifted_index(50000.0, snd_data["pres"], snd_data["vtmp"], vtmpk)

    assert (li == pytest.approx(-9.12, abs=1e-2))
    assert (pcl.pres == pytest.approx(92043.0))
    assert (pcl.tmpk == pytest.approx(298.15))
    assert (pcl.dwpk == pytest.approx(291.532))
    assert (pcl.cape == pytest.approx(3107.6, abs=5e-1))
    assert (pcl.cinh == pytest.approx(-36.4, abs=5e-1))
    assert (pcl.lfc_pressure == pytest.approx(71482, abs=1e0))
    assert (pcl.eql_pressure == pytest.approx(18969, abs=1e0))
    assert (pcl.mpl_pressure == pytest.approx(9529, abs=1e-0))


def test_downdraft_parcel():
    wobus = parcel.lifter_wobus()
    cm1 = parcel.lifter_cm1()
    cm1.ma_type = thermo.adiabat.pseudo_liq

    search_layer = layer.PressureLayer(
        snd_data["pres"][0],
        snd_data["pres"][0] - 40000.0
    )

    dcape_pcl = parcel.DowndraftParcel.min_thetae(
        search_layer,
        snd_data["pres"],
        snd_data["tmpk"],
        snd_data["dwpk"],
        snd_data["thetae"]
    )

    pcl_t_wobf = dcape_pcl.lower_parcel(wobus, snd_data["pres"])
    pcl_t_cm1_pseudo_liq = dcape_pcl.lower_parcel(cm1, snd_data["pres"])

    pcl_buoy_wobf = thermo.buoyancy(pcl_t_wobf, snd_data["tmpk"])
    pcl_buoy_cm1 = thermo.buoyancy(pcl_t_cm1_pseudo_liq, snd_data["tmpk"])

    dcape_wobf, dcinh_wobf = dcape_pcl.cape_cinh(
        snd_data["pres"], snd_data["hght"], pcl_buoy_wobf)
    dcape_cm1, dcinh_cm1 = dcape_pcl.cape_cinh(
        snd_data["pres"], snd_data["hght"], pcl_buoy_cm1)

    assert (dcape_wobf == pytest.approx(-1450.5, abs=2))
    assert (dcinh_wobf == 0.0)
    assert (dcape_cm1 == pytest.approx(-1422.1, abs=2))
    assert (dcinh_cm1 == 0.0)



from typing import Annotated

import numpy
from numpy.typing import NDArray


def interp_height(hght_val: float, hght_arr: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], data_arr: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> float:
    """
    Interpolate a value from an array in height coordinates (meters).
    The coordinate array (hght_arr) is assumed to be sorted (ascending).

    Parameters
    ----------
    hght_val : float 
        The coordinate height value to interpolate to (meters)
    hght_arr : numpy.ndarray[dtype=float32]
        1D numpy array of height values to interpolate from (meters)
    data_arr : numpy.ndarray[dtype=float32]
        float 1D numpy array of data values to interpolate from 

    Returns
    -------
    float 
        An interpolated data value from data_arr corresponding to hght_val
    """

def interp_pressure(pres_val: float, pres_arr: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], data_arr: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> float:
    """
    Interpolate a value from an array in pressure coordinates (Pa).
    All pressure interpolation happens in log10 space.
    The coordinate array (pres_arr) is assumed to be sorted (descending).

    Parameters
    ----------
    pres_val : float 
        The coordinate pressure value to interpolate to (Pa)
    pres_arr : nump.ndarray[dtype=float32] 
        1D numpy array of pressure values to interpolate from (Pa)
    data_arr : numpy.ndarray[dtype=float32]
        1D numpy array of data values to interpolate from 

    Returns
    -------
    float 
        An interpolated data value from data_arr corresponding to pres_val
    """

def find_first_pressure(data_val: float, pressure: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], data_array: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> float:
    """
    Conducts a bottom-up search for the first occurrence of a given value,
    and interpolates in order to get the pressure level it occurs at.

    Parameters
    ----------
    data_val : float 
        The value to search for 
    pressure : numpy.ndarray[dtype=float32]
        1D NumPy array of pressure values (Pa)
    data_array : numpy.ndarray[dtype=float32]
        1D NumPy array of values to search over

    Returns
    -------
    float
        The pressure level of first occurrence (Pa)
    """

def find_first_height(data_val: float, height: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)], data_array: Annotated[NDArray[numpy.float32], dict(shape=(None,), order='C', device='cpu', writable=False)]) -> float:
    """
    Conducts a bottom-up search for the first occurrence of a given value,
    and interpolates in order to get the height level it occurs at.

    Parameters
    ----------
    data_val : float 
        The value to search for 
    height : numpy.ndarray[dtype=float32]
        1D NumPy array of height values (meters)
    data_array : numpy.ndarray[dtype=float32]
        1D NumPy array of values to search over

    Returns
    -------
    float
        The height level of first occurrence (meters)
    """
